'''
Methods to create various types of Decoder objects from data(files)
'''
import re
import pickle
import sys

import numpy as np
from scipy.io import loadmat
from ..dio import parse

import tables
from . import kfdecoder, ppfdecoder
from . import state_space_models
import os

############
## Constants
############
pi = np.pi

################################################################################
## Functions to synchronize task-generated HDF files and neural recording files
################################################################################
def sys_eq(sys1, sys2):
    '''
    Determine if two system strings match. A separate function is required because sometimes
    the NIDAQ card doesn't properly transmit the first character of the name of the system.

    Parameters
    ----------
    sys1: string
        Name of system from the neural file
    sys2: string
        Name of system to match

    Returns
    -------
    Boolean indicating whether sys1 and sys2 match
    '''
    if sys2 == 'task':
        if sys1 in ['TAS\x00TASK', 'btqassskh', 'btqassskkkh', 'tasktasktask', 'task\x00task\x00task', b'task']:
            return True
        elif sys1[:4] in ['tqas', 'tacs','ttua', 'bttu', 'tttu']:
            return True

    return sys1 in [sys2, sys2[1:], sys2.upper()]


#FAKE_BLACKROCK_TMASK = True
FAKE_BLACKROCK_TMASK = False

########################################
## Neural data synchronization functions
########################################
def _get_tmask(files, tslice, sys_name='task'):
    if 'plexon' in files:
        fn = _get_tmask_plexon
        fname = str(files['plexon'])
    elif 'blackrock' in files:
        if FAKE_BLACKROCK_TMASK:
            fn = _get_tmask_blackrock_fake
            fname = files['hdf']
        else:
            fn = _get_tmask_blackrock
            fname = [str(name) for name in files['blackrock'] if '.nev' in name][0]  # only one of them
    elif 'ecube' in files:
        fn = _get_tmask_ecube
        fname = files
    else:
        raise Exception("Neural data file(s) not found!")

    return fn(fname, tslice, sys_name=sys_name)

def _get_tmask_plexon(plx, tslice, sys_name='task'):
    '''
    Find the rows of the plx file to use for training the decoder

    Parameters
    ----------
    plx : plexfile instance
        The plexon file to sync
    tslice : list of length 2
        Specify the start and end time to examine the file, in seconds
    sys_name : string, optional
        The "system" being synchronized. When the task is running, each data source
        (i.e., each HDF table) is allowed to be asynchronous and thus is independently
        synchronized with the neural recording system.

    Returns
    -------
    tmask: np.ndarray of shape (N, ) of booleans
        Specifies which entries of "rows" (see below) are within the time bounds
    rows: np.ndarray of shape (N, ) of integers
        The times at which rows of the specified HDF table were recieved in the neural recording box
    '''
    # Open plx file
    from plexon import plexfile
    if isinstance(plx, str) or isinstance(plx, str):
        plx = plexfile.openFile(plx.encode('utf-8'))

    # Get the list of all the systems registered in the neural data file
    events = plx.events[:].data
    reg = parse.registrations(events)

    if len(list(reg.keys())) > 0:
        # find the key for the specified system data
        syskey = None
        for key, system in list(reg.items()):
            if sys_eq(system[0], sys_name):
                syskey = key
                break

        if syskey is None:
            print((list(reg.items())))
            raise Exception('riglib.bmi.train._get_tmask: Training data source not found in neural data file!')
    elif len(list(reg.keys())) == 0:
        # try to find how many systems' rowbytes were in the HDF file
        rowbyte_data = parse.rowbyte(events)
        if len(list(rowbyte_data.keys())) == 1:
            print("No systems registered, but only one system registered with rowbytes! Using it anyway instead of throwing an error")
            syskey = list(rowbyte_data.keys())[0]
        else:
            raise Exception("No systems registered and I don't know which sys to use to train!")

    # get the corresponding hdf rows
    rows = parse.rowbyte(events)[syskey][:,0]

    # Determine which rows are within the time bounds
    lower, upper = 0 < rows, rows < rows.max() + 1
    l, u = tslice
    if l is not None:
        lower = l < rows
    if u is not None:
        upper = rows < u
    tmask = np.logical_and(lower, upper)
    return tmask, rows

def _get_tmask_blackrock(nev_fname, tslice, sys_name='task'):
    ''' Find the rows of the nev file to use for training the decoder.'''

    if nev_fname[-4:] != '.hdf':
        nev_hdf_fname = nev_fname + '.hdf'

        if not os.path.isfile(nev_hdf_fname):
            # convert .nev file to hdf file using our own blackrock_parse_files:
            from db.tracker import models
            task_entry = int(nev_fname[-8:-4])
            _, _ = models.parse_blackrock_file(nev_fname, 0, task_entry)
    else:
        nev_hdf_fname = nev_fname

    #import h5py
    #nev_hdf = h5py.File(nev_hdf_fname, 'r')
    nev_hdf = tables.openFile(nev_hdf_fname)

    #path = 'channel/digital00001/digital_set'
    #ts = nev_hdf.get(path).value['TimeStamp']
    #msgs = nev_hdf.get(path).value['Value']

    ts = nev_hdf.root.channel.digital0001.digital_set[:]['TimeStamp']
    msgs = nev_hdf.root.channel.digital0001.digital_set[:]['Value'] + 2**16

    msgtype = np.right_shift(np.bitwise_and(msgs, parse.msgtype_mask), 8).astype(np.uint8)
    # auxdata = np.right_shift(np.bitwise_and(msgs, auxdata_mask), 8).astype(np.uint8)
    auxdata = np.right_shift(np.bitwise_and(msgs, parse.auxdata_mask), 8+3).astype(np.uint8)
    rawdata = np.bitwise_and(msgs, parse.rawdata_mask)

    # data is an N x 4 matrix that will be the argument to parse.registrations()
    data = np.vstack([ts, msgtype, auxdata, rawdata]).T

    # get system registrations
    reg = parse.registrations(data)
    syskey = None

    for key, system in list(reg.items()):
            if sys_eq(system[0], sys_name):
                syskey = key
                break

    if syskey is None:
        raise Exception('No source registration saved in the file!')

    # get the corresponding hdf rows
    rows = parse.rowbyte(data)[syskey][:,0]

    rows = rows / 30000.

    lower, upper = 0 < rows, rows < rows.max() + 1
    if tslice is None:
        l = None;
        u = None;
    else:
        l, u = tslice

    if l is not None:
        lower = l < rows
    if u is not None:
        upper = rows < u
    tmask = np.logical_and(lower, upper)
    return tmask, rows

def _get_tmask_blackrock_fake(hdf_fname, tslice, **kwargs):
    # need to create fake "rows" and "tmask" variables

    print('WARNING: Using _get_tmask_blackrock_fake function!!')

    binlen = 0.1
    strobe_rate = 10
    hdf = tables.openFile(hdf_fname)

    n_rows = hdf.root.task[:]['plant_pos'].shape[0]
    first_ts = binlen
    rows = np.linspace(first_ts, first_ts + (n_rows-1)*(1./strobe_rate), num=n_rows)
    lower, upper = 0 < rows, rows < rows.max() + 1
    l, u = tslice
    if l is not None:
        lower = l < rows
    if u is not None:
        upper = rows < u
    tmask = np.logical_and(lower, upper)

    return tmask, rows

def _get_tmask_ecube(files, tslice, **kwargs):
    '''
    Find the "rows" of the ecube file to use for training the decoder.
    For compatibility with plexon and blackrock data, since actually we aren't saving
    any BMI3D rows into the ecube file, only sync'ing the cycle number via a strobe BNC
    output. 

    Args: 
        data_dir: location where the ecube data resides
        tslice : list of length 2
            Specify the start and end time to examine the file, in seconds

    Returns:
        tmask: np.ndarray of shape (N, ) of booleans
            Specifies which BMI3D cycles are within the time bounds
        rows: np.ndarray of shape (N, ) of integers
            The times at which BMI3D cycles were recorded on the ecube
    '''
    from riglib.ecube import load_bmi3d_cycle_times
    rows = load_bmi3d_cycle_times(files)

    # Determine which rows are within the time bounds
    lower, upper = 0 < rows, rows < rows.max() + 1
    l, u = tslice
    if l is not None:
        lower = l < rows
    if u is not None:
        upper = rows < u
    tmask = np.logical_and(lower, upper)
    return tmask, rows

################################################################################
## Feature extraction
################################################################################
def _get_neural_features_plx(files, binlen, extractor_fn, extractor_kwargs, tslice=None, units=None, source='task', strobe_rate=60.):
    '''
    Extract the neural features used to train the decoder

    Parameters
    ----------
    files: dict
        keys of the dictionary are file-types (e.g., hdf, plexon, etc.), values are file names
    binlen: float
        Specifies the temporal resolution of the feature extraction
    extractor_fn: callable
        Function must have the call signature
        neural_features, units, extractor_kwargs = extractor_fn(plx, neurows, binlen, units, extractor_kwargs)
    extractor_kwargs: dictionary
        Additional keyword arguments to the extractor_fn (specific to each feature extractor)

    Returns
    -------
    neural_features: np.ndarrya of shape (n_features, n_timepoints)
        Values of each feature to be used in training the decoder parameters
    units: np.ndarray of shape (N, -1)
        Specifies identty of each neural feature
    extractor_kwargs: dictionary
        Keyword arguments used to construct the feature extractor used online
    '''

    hdf = tables.open_file(files['hdf'])

    plx_fname = files['plexon'].encode('utf-8')
    from plexon import plexfile
    try:
        plx = plexfile.openFile(plx_fname)
    except IOError:
        raise Exception("Could not open .plx file: %s" % plx_fname)

    # Use all of the units if none are specified
    if units is None:
        units = np.array(plx.units).astype(np.int32)

    if tslice is None:
        tslice = (1., plx.length-1)

    tmask, rows = _get_tmask_plexon(plx, tslice, sys_name=source)
    neurows = rows[tmask]

    neural_features, units, extractor_kwargs = extractor_fn(files, neurows, binlen, units, extractor_kwargs)

    return neural_features, units, extractor_kwargs

def _get_neural_features_blackrock(files, binlen, extractor_fn, extractor_kwargs, tslice=None, units=None, source='task', strobe_rate=20.):
    if units is None:
        raise Exception('"units" variable is None in preprocess_files!')

    # Note: blackrock units are actually 0-based, but the units to be used for training
    #       (which comes from web interface) are 1-based; to account for this, add 1
    #       to unit numbers when reading from .nev file

    # notes:
    # tmask   --> logical vector of same length as rows that is True for rows inside the tslice
    # rows    --> times (in units of s, measured on neural system) that correspond to each row of the task in hdf file
    # kin     --> every 6th row of kinematics within the tslice boundaries
    # neurows --> the rows inside the tslice

    if FAKE_BLACKROCK_TMASK:
       tmask, rows = _get_tmask_blackrock_fake(files['hdf'], tslice)
    else:
        nev_fname = [name for name in files['blackrock'] if '.nev' in name][0]  # only one of them

        #tmask, rows = _get_tmask_blackrock(nev_fname, tslice, syskey_fn=lambda x: x[0] in [source, source[1:]])

        tmask, rows = _get_tmask_blackrock(nev_fname, tslice, sys_name=source)
    neurows = rows[tmask]

    neural_features, units, extractor_kwargs = extractor_fn(files, neurows, binlen, units, extractor_kwargs, strobe_rate=strobe_rate)

    return neural_features, units, extractor_kwargs

def _get_neural_features_tdt(files, binlen, extractor_fn, extractor_kwargs, tslice=None, units=None, source='task', strobe_rate=10.):
    raise NotImplementedError

def _get_neural_features_ecube(files, binlen, extractor_fn, extractor_kwargs, tslice=None, units=None, source='task', strobe_rate=120):
    '''
    Mostly just copied from _get_neural_features_plx()... see their docs
    '''
    from riglib.ecube import parse_file
    info = parse_file(files['ecube'])
    # Use all of the units if none are specified
    if units is None:
        units = np.array(info.units).astype(np.int32)

    if tslice is None:
        tslice = (1., info.length-1)

    tmask, rows = _get_tmask_ecube(files, tslice)
    neurows = rows[tmask]

    neural_features, units, extractor_kwargs = extractor_fn(files, neurows, binlen, units, extractor_kwargs, strobe_rate=strobe_rate)

    return neural_features, units, extractor_kwargs

def get_neural_features(files, binlen, extractor_fn, extractor_kwargs, units=None, tslice=None, source='task', strobe_rate=60):
    '''
    Docstring

    Parameters
    ----------

    Returns
    -------
    '''

    if 'plexon' in files:
        fn = _get_neural_features_plx
    elif 'blackrock' in files:
        fn = _get_neural_features_blackrock
        strobe_rate = 20.
    elif 'tdt' in files:
        fn = _get_neural_features_tdt
    elif 'ecube' in files:
        fn = _get_neural_features_ecube
    else:
        raise Exception('Could not find any recognized neural data files!')

    neural_features, units, extractor_kwargs = fn(files, binlen, extractor_fn, extractor_kwargs, tslice=tslice, units=units, source=source, strobe_rate=strobe_rate)

    return neural_features, units, extractor_kwargs

################################################################################
## Kinematic data retrieval
################################################################################
def null_kin_extractor(files, binlen, tmask, update_rate_hz=60., pos_key='cursor', vel_key=None):
    hdf = tables.openFile(files['hdf'])
    kin = np.squeeze(hdf.root.task[:][pos_key])

    inds, = np.nonzero(tmask)
    step_fl = binlen/(1./update_rate_hz)
    if step_fl < 1: # more than one spike bin per kinematic obs
        kin = np.hstack([kin, velocity])

        n_repeats = int((1./update_rate_hz)/binlen)
        inds = np.sort(np.hstack([inds]*n_repeats))
        kin = kin[inds]
    else:
        step = int(binlen/(1./update_rate_hz))
        inds = inds[::step]
        kin = kin[inds]

    print(("kin.shape", kin.shape))
    return kin


def get_plant_pos_vel(files, binlen, tmask, update_rate_hz=60., pos_key='cursor', vel_key=None):
    '''
    Get positions and velocity from 'task' table of HDF file

    Parameters
    ----------

    Returns
    -------
    '''
    if pos_key == 'plant_pos':  # used for ibmi tasks
        vel_key = 'plant_vel'

    hdf = tables.open_file(files['hdf'])
    kin = hdf.root.task[:][pos_key]

    inds, = np.nonzero(tmask)
    step_fl = binlen/(1./update_rate_hz)
    if step_fl < 1: # more than one spike bin per kinematic obs
        if vel_key is not None:
            velocity = hdf.root.task[:][vel_key]
        else:
            velocity = np.diff(kin, axis=0) * update_rate_hz
            velocity = np.vstack([np.zeros(kin.shape[1]), velocity])
        kin = np.hstack([kin, velocity])

        n_repeats = int((1./update_rate_hz)/binlen)
        inds = np.sort(np.hstack([inds]*n_repeats))
        kin = kin[inds]
    else:
        step = int(binlen/(1./update_rate_hz))
        inds = inds[::step]
        try:
            kin = kin[inds]
            if vel_key is not None:
                velocity = hdf.root.task[inds][vel_key]
            else:
                velocity = np.diff(kin, axis=0) * 1./binlen
                velocity = np.vstack([np.zeros(kin.shape[1]), velocity])

        except:
            kin2 = np.zeros((len(inds), kin.shape[1]))
            vel2 = np.zeros((len(inds), kin.shape[1]))

            ix = np.nonzero(inds < len(kin))[0]
            kin2[ix, :] = kin[inds[ix], :]
            kin = kin2.copy()

            if vel_key is not None:
                vel2[ix, :] = hdf.root.task[inds[ix]][vel_key]
            else:
                vel2 = np.diff(kin, axis=0) * 1./binlen
                vel2 = np.vstack([np.zeros(kin.shape[1]), vel2])

            velocity = vel2.copy()

        kin = np.hstack([kin, velocity])

    return kin


################################################################################
## Main training functions
################################################################################
def create_onedimLFP(files, extractor_cls, extractor_kwargs, kin_extractor, ssm, units, update_rate=0.1, tslice=None, kin_source='task',
    pos_key='cursor', vel_key=None, zscore=False):
    ## get neural features
    from . import extractor
    f_extractor = extractor.LFPMTMPowerExtractor(None, **extractor_kwargs)
    from . import onedim_lfp_decoder as old
    return old.create_decoder(units, ssm, extractor_cls, f_extractor.extractor_kwargs)

def test_ratBMIdecoder(te_id=None, update_rate=0.1, tslice=None, kin_source='task', pos_key='cursor', vel_key=None, **kwargs):
    from db import dbfunctions as dbfn
    te = dbfn.TaskEntry(te_id)
    files = dict(hdf=te.hdf_filename, plexon=te.plx_filename)

    entry = te.id
    from . import extractor
    extractor_cls = extractor.BinnedSpikeCountsExtractor

    neural_features, units, extractor_kwargs = get_neural_features(files, 0.1, extractor_cls.extract_from_file, dict(), tslice=None)
    extractor_kwargs['units'] = units

    from . import rat_bmi_decoder
    nsteps = kwargs.pop('nsteps', 10)
    prob_t1 = kwargs.pop('prob_t1', 0.985)
    prob_t2 = kwargs.pop('prob_t2', 0.015)
    timeout = kwargs.pop('timeout', 30.)
    timeout_pause = kwargs.pop('timeout_pause', 3.)
    freq_lim = kwargs.pop('freq_lim', (1000., 20000.))
    e1_inds = kwargs.pop('e1_inds', None)
    e2_inds = kwargs.pop('e2_inds', None)

    e1_inds, e2_inds, FR_to_freq_fn, units, t1, t2, mid = rat_bmi_decoder.calc_decoder_from_baseline_file(neural_features, units, nsteps, prob_t1, prob_t2, timeout,
        timeout_pause, freq_lim, e1_inds, e2_inds)

    task_params = dict(nsteps=nsteps, prob_t1=prob_t1, prob_t2=prob_t2, timeout_pause=timeout_pause, timeout=timeout, freq_lim=freq_lim,
        e1_inds=e1_inds, e2_inds=e2_inds, te_name=te.name, FR_to_freq_fn=FR_to_freq_fn, units=units, te_id=te_id, t1=t1, t2=t2, mid=mid,
        extractor_kwargs=extractor_kwargs)

    return task_params

def test_IsmoreSleepDecoder(te_id, e1_units, e2_units, nsteps=1, prob_t1 = 0.985, prob_t2 = 0.015, timeout = 15.,
    timeout_pause=0., freq_lim = [-1, 1], targets_matrix=None, session_length=0, saturate_perc=90,
    skip_sim=False):

    from db import dbfunctions as dbfn
    te = dbfn.TaskEntry(te_id)
    files = dict(hdf=te.hdf_filename, blackrock=te.blackrock_filenames)
    entry = te.id
    from . import extractor
    extractor_cls = extractor.BinnedSpikeCountsExtractor

    units = np.vstack((e1_units, e2_units))
    argsort = np.argsort(units[:, 0])
    units = units[argsort, :]

    unit_ids = np.hstack((['e1']*len(e1_units) + ['e2']*len(e2_units)))
    sorted_unit_ids = unit_ids[argsort]

    e1_inds = np.nonzero(sorted_unit_ids=='e1')[0]
    e2_inds = np.nonzero(sorted_unit_ids=='e2')[0]

    neural_features, units, extractor_kwargs = get_neural_features(files, 0.1, extractor_cls.extract_from_file,
        dict(), tslice=None, units=units)

    neural_features_unbinned, units, extractor_kwargs = get_neural_features(files, 0.05, extractor_cls.extract_from_file,
        dict(), tslice=None, units=units)

    import riglib.bmi.rat_bmi_decoder

    kwargs = dict(targets_matrix=targets_matrix, session_length=session_length,
        saturate_perc=saturate_perc, skip_sim=skip_sim)

    decoder, nrewards = riglib.bmi.rat_bmi_decoder.calc_decoder_from_baseline_file(neural_features,
        neural_features_unbinned, units, nsteps, prob_t1, prob_t2, timeout, timeout_pause, freq_lim,
        e1_inds, e2_inds, sim_fcn='ismore', **kwargs)

    decoder.extractor_cls = extractor_cls
    decoder.extractor_kwargs = extractor_kwargs
    pickle.dump(decoder, open('/storage/decoders/sleep_from_te'+str(te_id)+'.pkl', 'wb'))
    from db.tracker import dbq
    dbq.save_bmi('sleep_from_te'+str(te_id), te_id, '/storage/decoders/sleep_from_te'+str(te_id)+'.pkl')
    return decoder, nrewards

def create_ratBMIdecoder(task_params):
    from . import extractor
    task_params['extractor_cls'] = extractor.BinnedSpikeCountsExtractor
    from . import rat_bmi_decoder
    from . import state_space_models
    rat_decoder= rat_bmi_decoder.create_decoder(state_space_models.StateSpaceEndptPos1D(), task_params)
    rat_decoder.extractor_kwargs = task_params['extractor_kwargs']
    import tempfile
    import pickle
    from db.tracker import dbq

    rat_decoder.te_id = task_params['te_id']
    tf = tempfile.NamedTemporaryFile('wb')
    pickle.dump(rat_decoder, tf, 2)
    tf.flush()

    name = task_params['te_name'] + '_rat_bmi_decoder'
    dbq.save_bmi(name, int(task_params['te_id']), tf.name)

def create_lindecoder(files, extractor_cls, extractor_kwargs, kin_extractor, ssm, units=None, update_rate=0.1, tslice=None, kin_source='task',
    pos_key='cursor', vel_key=None, zscore=False):
    from . import lindecoder, state_space_models
    
    # Hack job incoming:
    if 'mouse' in files:
        neural_data = [[0., 0.], [1650., 1080.]]
        units = [(1, 0), (2, 0)]
        unit_to_state = None
        decoder_to_plant = 20
        smoothing_window = 1
        vel_control = False
    return lindecoder.create_lindecoder(ssm, units, neural_data, unit_to_state, decoder_to_plant, smoothing_window, vel_control, update_rate)

def add_fa_dict_to_decoder(decoder_training_te, dec_ix, fa_te):
    #First make sure we're training from the correct task entry: spike counts n_units == BMI units
    from db import dbfunctions as dbfn
    te = dbfn.TaskEntry(fa_te)
    hdf = te.hdf
    sc_n_units = hdf.root.task[0]['spike_counts'].shape[0]


    from db.tracker import models
    te_arr = models.Decoder.objects.filter(entry=decoder_training_te)
    search_flag = 1
    for te in te_arr:
        ix = te.path.find('_')
        if search_flag:
            if int(te.path[ix+1:ix+3]) == dec_ix:
                decoder_old = te
                search_flag = 0

    if search_flag:
        raise Exception('No decoder from ', str(decoder_training_te), ' and matching index: ', str(dec_ix))

    from tasks.factor_analysis_tasks import FactorBMIBase
    FA_dict = FactorBMIBase.generate_FA_matrices(fa_te)

    import pickle
    dec = pickle.load(open(decoder_old.filename))
    dec.trained_fa_dict = FA_dict
    dec_n_units = dec.n_units

    if dec_n_units != sc_n_units:
        raise Exception('Cant use TE for BMI training and FA training -- n_units mismatch')

    from db import trainbmi
    trainbmi.save_new_decoder_from_existing(dec, decoder_old, suffix='_w_fa_dict_from_'+str(fa_te))

def train_FADecoder_from_KF(FA_nfactors, FA_te_id, decoder, use_scaled=True, use_main=True):

    from tasks.factor_analysis_tasks import FactorBMIBase
    FA_dict = FactorBMIBase.generate_FA_matrices(FA_nfactors, FA_te_id)

    # #Now, retrain:
    binlen = decoder.binlen

    from db import dbfunctions as dbfn
    te_id = dbfn.TaskEntry(decoder.te_id)
    files = dict(plexon=te_id.plx_filename, hdf = te_id.hdf_filename)
    extractor_cls = decoder.extractor_cls
    extractor_kwargs = decoder.extractor_kwargs
    kin_extractor = get_plant_pos_vel
    ssm = decoder.ssm
    update_rate = decoder.binlen
    units = decoder.units
    tslice = (0., te_id.length)

    ## get kinematic data
    kin_source = 'task'
    tmask, rows = _get_tmask(files, tslice, sys_name=kin_source)
    kin = kin_extractor(files, binlen, tmask, pos_key='cursor', vel_key=None)

    ## get neural features
    neural_features, units, extractor_kwargs = get_neural_features(files, binlen, extractor_cls.extract_from_file, extractor_kwargs, tslice=tslice, units=units, source=kin_source)

    #Get shared input:
    T = neural_features.shape[0]
    demean = neural_features.T - np.tile(FA_dict['fa_mu'], [1, T])

    if use_main:
        main_shar = (FA_dict['fa_main_shared'] * demean)
        main_priv = (demean - main_shar)
        FA = FA_dict['FA_model']

    else:
        shar = (FA_dict['fa_sharL']* demean)
        shar_sc = np.multiply(shar, np.tile(FA_dict['fa_shar_var_sc'], [1, T])) + np.tile(FA_dict['fa_mu'], [1, T])
        shar_unsc = shar + np.tile(FA_dict['fa_mu'], [1, T])
        if use_scaled:
            neural_features = shar_sc[:,:-1]
        else:
            neural_features = shar_unsc[:,:-1]

    # Remove 1st kinematic sample and last neural features sample to align the
    # velocity with the neural features
    kin = kin[1:].T

    decoder2 = train_KFDecoder_abstract(ssm, kin, neural_features, units, update_rate, tslice=tslice)
    decoder2.extractor_cls = extractor_cls
    decoder2.extractor_kwargs = extractor_kwargs
    decoder2.te_id = decoder.te_id
    decoder2.trained_fa_dict = FA_dict

    import datetime
    now = datetime.datetime.now()
    tp = now.isoformat()
    import pickle
    fname = os.path.expandvars('$FA_GROM_DATA/decoder_')+tp+'.pkl'
    f = open(fname, 'w')
    pickle.dump(decoder2, f)
    f.close()
    return decoder2, fname

def conv_KF_to_splitFA_dec(decoder_training_te, dec_ix, fa_te, search_suffix = 'w_fa_dict_from_', use_shar_z=False, tslice=None):

    from db import dbfunctions as dbfn
    te = dbfn.TaskEntry(fa_te)
    hdf = te.hdf
    sc_n_units = hdf.root.task[0]['spike_counts'].shape[0]

    from db.tracker import models
    te_arr = models.Decoder.objects.filter(entry=decoder_training_te)
    search_flag = 1
    for te in te_arr:
        ix = te.path.find('_')
        if search_flag:
            if int(te.path[ix+1:ix+3]) == dec_ix:
                decoder = pickle.load(open(te.filename))
                if hasattr(decoder, 'trained_fa_dict'):
                    ix = te.path.find('w_fa_dict_from_')
                    if ix > 1:
                        fa_te_train = te.path[ix+len(search_suffix):ix+len(search_suffix)+4]
                        if int(fa_te_train) == fa_te:
                            decoder_old = te
                            #search_flag = 0

    # if search_flag:
    #     raise Exception('No decoder from ', str(decoder_training_te), ' and matching index: ', str(dec_ix), ' with FA training from: ',str(fa_te))
    # else:
    print(('Using old decoder: ', decoder_old.path))

    decoder = pickle.load(open(decoder_old.filename))
    if hasattr(decoder, 'trained_fa_dict'):
        FA_dict = decoder.trained_fa_dict
    else:
        raise Exception('Make an FA dict decoder first, then re-train that')

    from db import dbfunctions as dbfn
    te_id = dbfn.TaskEntry(fa_te)

    files = dict(plexon=te_id.plx_filename, hdf = te_id.hdf_filename)
    extractor_cls = decoder.extractor_cls
    extractor_kwargs = decoder.extractor_kwargs
    extractor_kwargs['discard_zero_units'] = False
    kin_extractor = get_plant_pos_vel
    ssm = decoder.ssm
    update_rate = binlen = decoder.binlen
    units = decoder.units
    if tslice is None:
        tslice = (0., te_id.length)

    ## get kinematic data
    kin_source = 'task'
    tmask, rows =_get_tmask(files, tslice, sys_name=kin_source)
    kin = kin_extractor(files, binlen, tmask, pos_key='cursor', vel_key=None)

    ## get neural features
    neural_features, units, extractor_kwargs = get_neural_features(files, binlen, extractor_cls.extract_from_file, extractor_kwargs, tslice=tslice, units=units, source=kin_source)

    #Get main shared input:
    T = neural_features.shape[0]
    demean = neural_features.T - np.tile(FA_dict['fa_mu'], [1, T])

    #Neural features in time x spikes:
    FA = FA_dict['FA_model']
    z = FA.transform(demean.T)
    z = z.T
    z = z[:FA_dict['fa_main_shar_n_dim'], :]

    #z = FA_dict['u_svd'].T*FA_dict['uut_psi_inv']*demean

    shar_z = FA_dict['fa_main_shared'] * demean
    priv = demean - shar_z

    #Time by features:
    if use_shar_z:
        neural_features2 = np.vstack((z, priv))
        suffx = '_split_shar_z'
    else:
        neural_features2 = np.vstack((z, priv))
        suffx = '_split_z'
    decoder_split = train_KFDecoder_abstract(ssm, kin.T, neural_features2, units, update_rate, tslice=tslice)
    decoder_split.n_features = len(units)
    decoder_split.trained_fa_dict = FA_dict

    decoder_split.extractor_cls = extractor_cls
    decoder_split.extractor_kwargs = extractor_kwargs

    from db import trainbmi
    trainbmi.save_new_decoder_from_existing(decoder_split, decoder_old, suffix=suffx)

def train_KFDecoder(files, extractor_cls, extractor_kwargs, kin_extractor, ssm, units, update_rate=0.1, tslice=None,
    kin_source='task', pos_key='cursor', vel_key=None, zscore=False, filter_kin=False, simple_lin_reg=False,
    use_data_kwargs=None, update_rate_hz=60, **kwargs):
    '''
    Create a new KFDecoder using maximum-likelihood, from kinematic observations and neural observations

    Parameters
    ----------
    files : dict
        Dictionary of files which contain training data. Keys are file tyes, values are file names.
        Kinematic data is assumed to be stored in an 'hdf' file and neural data assumed to be in 'plx' or 'nev' files
    extractor_cls : class
        Class of feature extractor to instantiate
    extractor_kwargs : dict
        Parameters to specify for feature extractor to instantiate it to specification
    kin_extractor : callable
        Function to extract kinematics from the HDF file.
    ssm : state_space_models.StateSpace instance
        State space model for the Decoder object being created.
    units : np.iterable
        Spiking units are specified as tuples of (electrode channe, electrode unit)
    update_rate : float, optional
        Time in seconds between decoder updates. default=0.1
    tslice : iterable of length 2, optional
        Start and end times in seconds to specify the portion of the training data to use for ML estimation. By default, the whole dataset will be used
    kin_source : string, optional
        Table from the HDF file to grab kinematic data. Default is the 'task' table.
    pos_key : string, optional
        Column of HDF table to use for position data. Default is 'cursor', recognized options are {'cursor', 'joint_angles', 'plant_pos'}
    vel_key : string
        Column of HDF table to use for velocity data. Default is None; velocity is computed by single-step numerical differencing (or alternate method )
    zscore : Bool
        Determines whether to zscore neural_data or not
    kwargs:
        mFR: mean firing rate to use to zscore units
        sdFR: standard dev. to use to zscore units

    Returns
    -------
    KFDecoder instance
    '''
    import sys
    print(files)
    # sys.stdout.write(files)
    # sys.stdout.write(extractor_cls)
    # sys.stdout.write(extractor_kwargs.keys())
    # sys.stdout.write(units)
    binlen = update_rate

    ## get kinematic data
    tmask, rows = _get_tmask(files, tslice, sys_name=kin_source)
    kin = kin_extractor(files, binlen, tmask, pos_key=pos_key, vel_key=vel_key, update_rate_hz=update_rate_hz)

    ## get neural features
    if 'blackrock' in list(files.keys()):
        strobe_rate = 20.
    elif 'ecube' in list(files.keys()):
        strobe_rate = update_rate_hz # they are always the same
    else:
        strobe_rate = 60.
    
    neural_features, units, extractor_kwargs = get_neural_features(files, binlen, extractor_cls.extract_from_file,
        extractor_kwargs, tslice=tslice, units=units, source=kin_source, strobe_rate=strobe_rate)

    # Remove 1st kinematic sample and last neural features sample to align the
    # velocity with the neural features
    kin = kin[1:].T
    neural_features = neural_features[:-1].T

    if filter_kin:
        n_channels = len(kin)
        filts = get_filterbank(fs=1./update_rate, n_channels=n_channels)
        kin_filt = np.zeros_like(kin)
        for chan in range(n_channels):
            for filt in filts[chan]:
                kin_filt[chan, :] = filt(kin[chan, :])
    else:
        kin_filt = kin.copy()

    if simple_lin_reg:
        from sklearn.linear_model import Ridge
        decoder = Ridge(1000.0, fit_intercept=True, normalize=False)

        if use_data_kwargs is not None:

            # HDF rows to use in training
            X = []
            Y = []

            for pair in use_data_kwargs['pairs']:
                X.append(neural_features[:, pair[0]])
                Y.append(kin_filt[:, pair[1]])

            # Convert these hdf rows to
        decoder.fit(np.vstack((X)), np.vstack((Y)))

    else:
        decoder = train_KFDecoder_abstract(ssm, kin_filt, neural_features, units, update_rate, tslice=tslice, zscore=zscore, **kwargs)
        decoder.extractor_cls = extractor_cls
        decoder.extractor_kwargs = extractor_kwargs

    return decoder #, neural_features, kin_filt

def get_filterbank(n_channels=14, fs=1000.):
    # from ismore.filter import Filter
    from scipy.signal import butter, filtfilt
    band  = [.001, 1]  # Hz
    nyq   = 0.5 * fs
    low   = band[0] / nyq
    high  = band[1] / nyq
    high = np.min([high, 0.99])
    bpf_coeffs = butter(4, [low, high], btype='band')

    channel_filterbank = [None]*n_channels
    for k in range(n_channels):
        # filts = [Filter(bpf_coeffs[0], bpf_coeffs[1])]
        filts = [lambda x: filtfilt(bpf_coeffs[0], bpf_coeffs[1], x)]
        channel_filterbank[k] = filts
    return channel_filterbank


def train_KFDecoderDrift(files, extractor_cls, extractor_kwargs, kin_extractor, ssm, units, update_rate=0.1, tslice=None,
    kin_source='task', pos_key='cursor', vel_key=None, zscore=False, update_rae_hz=60, **kwargs):
    '''
    Create a new KFDecoder using maximum-likelihood, from kinematic observations and neural observations

    Parameters
    ----------
    files : dict
        Dictionary of files which contain training data. Keys are file tyes, values are file names.
        Kinematic data is assumed to be stored in an 'hdf' file and neural data assumed to be in 'plx' or 'nev' files
    extractor_cls : class
        Class of feature extractor to instantiate
    extractor_kwargs : dict
        Parameters to specify for feature extractor to instantiate it to specification
    kin_extractor : callable
        Function to extract kinematics from the HDF file.
    ssm : state_space_models.StateSpace instance
        State space model for the Decoder object being created.
    units : np.iterable
        Spiking units are specified as tuples of (electrode channe, electrode unit)
    update_rate : float, optional
        Time in seconds between decoder updates. default=0.1
    tslice : iterable of length 2, optional
        Start and end times in seconds to specify the portion of the training data to use for ML estimation. By default, the whole dataset will be used
    kin_source : string, optional
        Table from the HDF file to grab kinematic data. Default is the 'task' table.
    pos_key : string, optional
        Column of HDF table to use for position data. Default is 'cursor', recognized options are {'cursor', 'joint_angles', 'plant_pos'}
    vel_key : string
        Column of HDF table to use for velocity data. Default is None; velocity is computed by single-step numerical differencing (or alternate method )
    zscore : Bool
        Determines whether to zscore neural_data or not
    kwargs:
        mFR: mean firing rate to use to zscore units
        sdFR: standard dev. to use to zscore units

    Returns
    -------
    KFDecoder instance
    '''
    import sys
    print(files)
    # sys.stdout.write(files)
    # sys.stdout.write(extractor_cls)
    # sys.stdout.write(extractor_kwargs.keys())
    # sys.stdout.write(units)
    binlen = update_rate

    ## get kinematic data
    tmask, rows = _get_tmask(files, tslice, sys_name=kin_source)
    kin = kin_extractor(files, binlen, tmask, pos_key=pos_key, vel_key=vel_key, update_rate_hz=update_rate_hz)

    ## get neural features
    neural_features, units, extractor_kwargs = get_neural_features(files, binlen, extractor_cls.extract_from_file,
        extractor_kwargs, tslice=tslice, units=units, source=kin_source)

    # Remove 1st kinematic sample and last neural features sample to align the
    # velocity with the neural features
    kin = kin[1:].T
    neural_features = neural_features[:-1].T

    kwargs['driftKF'] = True
    decoder = train_KFDecoder_abstract(ssm, kin, neural_features, units, update_rate,
        tslice=tslice, zscore=zscore, **kwargs)

    decoder.extractor_cls = extractor_cls
    decoder.extractor_kwargs = extractor_kwargs

    return decoder

def train_KFDecoder_abstract(ssm, kin, neural_features, units, update_rate, tslice=None, regularizer=0.,
    zscore=False, **kwargs):
    print(kwargs)
    print('end of kwargs')

    #### Train the actual KF decoder matrices ####
    if type(zscore) is bool:
        pass
    else:
        if zscore == 'on':
            zscore = True
        elif zscore == 'off':
            zscore = False
        else:
            raise Exception

    print(('zscore value: ', zscore, type(zscore)))

    if zscore:
        if 'mFR' in kwargs and 'sdFR' in kwargs:
            print('using kwargs mFR, sdFR to zscore')
            mFR = kwargs['mFR']
            sdFR = kwargs['sdFR']
        else:
            print('computing own mFR, sdFR to zscore')
            mFR = np.mean(neural_features, axis=1)
            sdFR = np.std(neural_features, axis=1)
            if hasattr(kwargs, 'zscore_set_std_to_one'):
                sdFR = np.ones_like(mFR)
        neural_features = (neural_features - mFR[:, np.newaxis])*(1./sdFR[:, np.newaxis])

    else:
        mFR = np.squeeze(np.mean(neural_features, axis=1))
        sdFR = np.squeeze(np.std(neural_features, axis=1))

    if 'noise_rej' in kwargs:
        if kwargs['noise_rej']:
            sum_pop = np.sum(neural_features, axis = 0)
            bins_noisy = np.nonzero(sum_pop > kwargs['noise_rej_cutoff'])[0]
            print(('replacing %d noisy bins of total %d bins w/ mFR for decoder training!' % (len(bins_noisy), len(sum_pop))))
            neural_features[:, bins_noisy] = mFR[:, np.newaxis]
    else:
        kwargs['noise_rej'] = False
        kwargs['noise_rej_cutoff'] = -1.

    n_features = len(mFR)

    # C should be trained on all of the stochastic state variables, excluding the offset terms
    C = np.zeros((n_features, ssm.n_states))
    C[:, ssm.drives_obs_inds], Q = kfdecoder.KalmanFilter.MLE_obs_model(kin[ssm.train_inds, :], neural_features, regularizer=regularizer)


    # Set state space model
    A, B, W = ssm.get_ssm_matrices(update_rate=update_rate)

    # instantiate KFdecoder
    driftKF = kwargs.pop('driftKF', False)
    if driftKF:
        print(('Training Drift Decoder. Noise Rejection? ', kwargs['noise_rej']))
        kf = kfdecoder.KalmanFilterDriftCorrection(A, W, C, Q, is_stochastic=ssm.is_stochastic)
    else:
        kf = kfdecoder.KalmanFilter(A, W, C, Q, is_stochastic=ssm.is_stochastic)

    decoder = kfdecoder.KFDecoder(kf, units, ssm, binlen=update_rate, tslice=tslice)

    if zscore:
        decoder.init_zscore(mFR, sdFR)
        print('zscore init')
    else:
        print('no init_zscore')


    # Compute sufficient stats for C and Q matrices (used for RML CLDA)
    from .clda import KFRML
    n_features, n_states = C.shape
    R = np.mat(np.zeros([n_states, n_states]))
    S = np.mat(np.zeros([n_features, n_states]))
    R_small, S_small, T, ESS = KFRML.compute_suff_stats(kin[ssm.train_inds, :], neural_features)

    R[np.ix_(ssm.drives_obs_inds, ssm.drives_obs_inds)] = R_small
    S[:,ssm.drives_obs_inds] = S_small

    decoder.filt.R = R
    decoder.filt.S = S
    decoder.filt.T = T
    decoder.filt.ESS = ESS
    decoder.n_features = n_features

    decoder.filt.noise_rej = kwargs['noise_rej']
    decoder.filt.noise_rej_cutoff = kwargs['noise_rej_cutoff']
    decoder.filt.noise_rej_mFR = mFR
    # decoder.extractor_cls = extractor_cls
    # decoder.extractor_kwargs = extractor_kwargs

    return decoder

def train_PPFDecoder(files, extractor_cls, extractor_kwargs, kin_extractor, ssm, units, update_rate=0.1, tslice=None, kin_source='task',
    pos_key='cursor', vel_key=None, zscore=False):
    '''
    Create a new PPFDecoder using maximum-likelihood, from kinematic observations and neural observations

    Parameters
    ----------
    files : dict
        Dictionary of files which contain training data. Keys are file tyes, values are file names.
        Kinematic data is assumed to be stored in an 'hdf' file and neural data assumed to be in 'plx' or 'nev' files
    extractor_cls : class
        Class of feature extractor to instantiate
    extractor_kwargs : dict
        Parameters to specify for feature extractor to instantiate it to specification
    kin_extractor : callable
        Function to extract kinematics from the HDF file.
    ssm : state_space_models.StateSpace instance
        State space model for the Decoder object being created.
    units : np.iterable
        Spiking units are specified as tuples of (electrode channe, electrode unit)
    update_rate : float, optional
        Time in seconds between decoder updates. default=0.1
    tslice : iterable of length 2, optional
        Start and end times in seconds to specify the portion of the training data to use for ML estimation. By default, the whole dataset will be used
    kin_source : string, optional
        Table from the HDF file to grab kinematic data. Default is the 'task' table.
    pos_key : string, optional
        Column of HDF table to use for position data. Default is 'cursor', recognized options are {'cursor', 'joint_angles', 'plant_pos'}
    vel_key : string
        Column of HDF table to use for velocity data. Default is None; velocity is computed by single-step numerical differencing (or alternate method )

    Returns
    -------
    PPFDecoder instance
    '''
    binlen = 1./180 #update_rate

    ## get kinematic data
    tmask, rows = _get_tmask(files, tslice, sys_name=kin_source)
    kin = kin_extractor(files, binlen, tmask, pos_key=pos_key, vel_key=vel_key)

    ## get neural features
    neural_features, units, extractor_kwargs = get_neural_features(files, binlen, extractor_cls.extract_from_file, extractor_kwargs, tslice=tslice, units=units, source=kin_source)

    # Remove 1st kinematic sample and last neural features sample to align the
    # velocity with the neural features
    kin = kin[1:].T
    neural_features = neural_features[:-1].T

    decoder = train_PPFDecoder_abstract(ssm, kin, neural_features, units, update_rate, tslice=tslice)

    decoder.extractor_cls = extractor_cls
    decoder.extractor_kwargs = extractor_kwargs

    return decoder

def train_PPFDecoder_abstract(ssm, kin, neural_features, units, update_rate, tslice=None):
    binlen = 1./180 #update_rate
    # squash any spike counts greater than 1 (doesn't work with PPF model)
    neural_features[neural_features > 1] = 1

    #### Train the  PPF decoder matrices ####
    n_features = neural_features.shape[0]  # number of neural features

    # C should be trained on all of the stochastic state variables, excluding the offset terms
    C = np.zeros([n_features, ssm.n_states])
    C[:, ssm.drives_obs_inds], pvals = ppfdecoder.PointProcessFilter.MLE_obs_model(kin[ssm.train_inds, :], neural_features)

    # Set state space model
    A, B, W = ssm.get_ssm_matrices(update_rate=update_rate)

    # instantiate Decoder
    ppf = ppfdecoder.PointProcessFilter(A, W, C, B=B, dt=update_rate, is_stochastic=ssm.is_stochastic)
    decoder = ppfdecoder.PPFDecoder(ppf, units, ssm, binlen=binlen, tslice=tslice)

    # Compute sufficient stats for C matrix (used for RML CLDA)
    from .clda import KFRML
    n_features, n_states = C.shape
    S = np.mat(np.zeros([n_features, n_states]))
    S_small, = decoder.compute_suff_stats(kin[ssm.train_inds, :], neural_features)

    S[:,ssm.drives_obs_inds] = S_small

    decoder.filt.S = S
    decoder.n_features = n_features

    return decoder

###################
## Helper functions
###################
def unit_conv(starting_unit, ending_unit):
    '''
    Convert between units, e.g. cm to m
    Lookup table for conversion factors between units; this function exists
    only to avoid hard-coded constants in most of the code

    Parameters
    ----------
    starting_unit : string
        Name of current unit for the quantity, e.g., 'cm'
    ending_unit : string
        Name of desired unit for the quantity, e.g., 'm'

    Returns
    -------
    float
        Multiplicative scale factor to convert a scalar in the 'starting_unit' to the 'ending_unit'
    '''

    if starting_unit == ending_unit:
        return 1
    elif (starting_unit, ending_unit) == ('cm', 'm'):
        return 0.01
    elif (starting_unit, ending_unit) == ('m', 'cm'):
        return 100
    else:
        raise ValueError("Unrecognized starting/ending unit")

def lookup_cells(cells):
    '''
    Convert string names of units to 'machine' format.
    Take a list of neural units specified as a list of strings and convert
    to the 2D array format used to specify neural units to train decoders

    Parameters
    ----------
    cells : string
        String of cell names to parse, e.g., '1a, 2b'

    Returns
    -------
    list of 2-tuples
        Each element of the list is a tuple of (channel, unit), e.g., [(1, 1), (2, 2)]
    '''
    cellname = re.compile(r'(\d{1,3})\s*(\w{1})')
    cells = [ (int(c), ord(u) - 96) for c, u in cellname.findall(cells)]
    return cells

def inflate(A, current_states, full_state_ls, axis=0):
    '''
    'Inflate' a matrix by filling in rows/columns with zeros

    Docstring

    Parameters
    ----------

    Returns
    -------

    '''
    try:
        nS = len(full_state_ls)
    except:
        nS = full_state_ls.n_states

    if axis == 0:
        A_new = np.zeros([nS, A.shape[1]])
    elif axis == 1:
        A_new = np.zeros([A.shape[0], nS])

    try:
        new_inds = [full_state_ls.index(x) for x in current_states]
    except:
        new_inds = [full_state_ls.state_names.index(x) for x in current_states]
    if axis == 0:
        A_new[new_inds, :] = A
    elif axis == 1:
        A_new[:, new_inds] = A

    return A_new


#######################
## Simulation functions
#######################
def _train_PPFDecoder_2D_sim(stochastic_states, neuron_driving_states, units,
    bounding_box, states_to_bound, include_y=True, dt=0.1, v=0.4):
    '''
    Train a simulation PPFDecoder

    Docstring

    Parameters
    ----------

    Returns
    -------
    '''
    raise NotImplementedError

def rand_KFDecoder(ssm, units, dt=0.1):
    '''
    Make a KFDecoder with the observation model initialized randomly

    Parameters
    ----------
    ssm : state_space_models.StateSpace instance
        State-space model for the KFDecoder. Should specify the A and W matrices
    units : np.array of shape (N, 2)
        Unit labels to assign to each row of the C matrix

    Returns
    -------
    KFDecoder instance
    '''
    n_neurons = units.shape[0]
    binlen = dt

    A, B, W = ssm.get_ssm_matrices(update_rate=dt)
    drives_neurons = ssm.drives_obs
    is_stochastic = ssm.is_stochastic
    nX = ssm.n_states

    C = np.random.standard_normal([n_neurons, nX])
    C[:, ~drives_neurons] = 0
    Q = 10 * np.identity(n_neurons)

    kf = kfdecoder.KalmanFilter(A, W, C, Q, is_stochastic=is_stochastic)

    mFR = 0
    sdFR = 1
    decoder = kfdecoder.KFDecoder(kf, units, ssm, mFR=mFR, sdFR=sdFR, binlen=binlen)

    decoder.kf.R = np.mat(np.identity(decoder.kf.C.shape[1]))
    decoder.kf.S = decoder.kf.C
    decoder.kf.T = decoder.kf.Q + decoder.kf.S*decoder.kf.S.T
    decoder.kf.ESS = 3000.

    decoder.ssm = ssm
    decoder.n_features = n_neurons

    # decoder.bounder = make_rect_bounder_from_ssm(ssm)

    return decoder

_train_KFDecoder_2D_sim_2 = rand_KFDecoder

def make_fixed_kf_decoder(units, ssm, C, dt=0.1):
    n_neurons = units.shape[0]
    assert n_neurons == C.shape[0], "C matrix must have same first dimension as number of neurons"
    binlen = dt

    A, B, W = ssm.get_ssm_matrices(update_rate=dt)
    drives_neurons = ssm.drives_obs
    is_stochastic = ssm.is_stochastic
    nX = ssm.n_states
    assert nX == C.shape[1], "C matrix must have same second dimension as number of states"

    Q = 10 * np.identity(n_neurons)

    kf = kfdecoder.KalmanFilter(A, W, C, Q, is_stochastic=is_stochastic)

    mFR = 0
    sdFR = 1
    decoder = kfdecoder.KFDecoder(kf, units, ssm, mFR=mFR, sdFR=sdFR, binlen=binlen)

    decoder.kf.R = np.mat(np.identity(decoder.kf.C.shape[1]))
    decoder.kf.S = decoder.kf.C
    decoder.kf.T = decoder.kf.Q + decoder.kf.S*decoder.kf.S.T
    decoder.kf.ESS = 3000.

    decoder.ssm = ssm
    decoder.n_features = n_neurons

    # decoder.bounder = make_rect_bounder_from_ssm(ssm)

    return decoder

def load_from_mat_file(decoder_fname, bounding_box=None,
    states=['p_x', 'p_y', 'v_x', 'v_y', 'off'], states_to_bound=[]):
    """
    Create KFDecoder from MATLAB decoder file used in a Dexterit-based
    BMI

    Docstring

    Parameters
    ----------

    Returns
    -------
    """
    decoder_data = loadmat(decoder_fname)['decoder']
    A = decoder_data['A'][0,0]
    W = decoder_data['W'][0,0]
    H = decoder_data['H'][0,0]
    Q = decoder_data['Q'][0,0]
    mFR = decoder_data['mFR'][0,0].ravel()
    sdFR = decoder_data['sdFR'][0,0].ravel()

    pred_sigs = [str(x[0]) for x in decoder_data['predSig'][0,0].ravel()]
    unit_lut = {'a':1, 'b':2, 'c':3, 'd':4}
    units = [(int(sig[3:6]), unit_lut[sig[-1]]) for sig in pred_sigs]

    drives_neurons = np.array([False, False, True, True, True])

    kf = kfdecoder.KalmanFilter(A, W, H, Q)
    dec = kfdecoder.KFDecoder(kf, mFR, sdFR, units, bounding_box, states, drives_neurons, states_to_bound)

    # Load bounder for position state
    from state_bounders import RectangularBounder
    bounding_box_data = loadmat('/Users/sgowda/bmi/workspace/decoder_switching/jeev_center_out_bmi_targets.mat')
    center_pos = bounding_box_data['centerPos'].ravel()
    px_min, py_min = center_pos - 0.09
    px_max, py_max = center_pos + 0.09
    bounding_box = [(px_min, px_max), (py_min, py_max)]
    bounder = RectangularBounder([px_min, py_min], [px_max, py_max], ['p_x', 'p_y'])
    dec.bounder = bounder

    return dec

def rescale_KFDecoder_units(dec, scale_factor=10):
    '''
    Convert the units of a KFDecoder, e.g. from mm to cm

    C and W matrices of KalmanFilter must be updated for the new units.
    A and Q are unitless and thus remain the same

    Parameters
    ----------
    dec : KFDecoder instance
        KFDecoder object
    scale_factor : numerical
        defines how much bigger the new unit is than the old one
    '''
    inds = np.nonzero((np.diag(dec.kf.W) > 0) * dec.drives_neurons)[0]
    nS = dec.kf.W.shape[0]
    S_diag = np.ones(nS)
    S_diag[inds] = scale_factor
    S = np.mat(np.diag(S_diag))
    #S = np.mat(np.diag([1., 1, 1, 10, 10, 10, 1]))
    dec.kf.C *= S
    dec.kf.W *= S.I * S.I
    try:
        dec.kf.C_xpose_Q_inv_C = S.T * dec.kf.C_xpose_Q_inv_C * S
        dec.kf.C_xpose_Q_inv = S.T * dec.kf.C_xpose_Q_inv
    except:
        pass
    dec.bounding_box = tuple([x / scale_factor for x in dec.bounding_box])
    return dec

def _train_PPFDecoder_sim_known_beta(beta, units, dt=0.005, dist_units='m'):
    '''
    Create a PPFDecoder object to decode 2D velocity from a known 'beta' matrix

    Docstring

    Parameters
    ----------

    Returns
    -------
    '''
    units_mult_lut = dict(m=1., cm=0.01)
    units_mult = units_mult_lut[dist_units]

    ssm = state_space_models.StateSpaceEndptVel2D()
    A, _, W = ssm.get_ssm_matrices(update_rate=dt)

    # rescale beta for units
    beta[:,3:6] *= units_mult

    # Control input matrix for SSM for control inputs
    I = np.mat(np.eye(3))
    B = np.vstack([0*I, dt*1000 * I, np.zeros([1,3])])

    # instantiate Decoder
    ppf = ppfdecoder.PointProcessFilter(A, W, beta, dt=dt, is_stochastic=ssm.is_stochastic, B=B)
    dec = ppfdecoder.PPFDecoder(ppf, units, ssm, binlen=dt)

    n_stoch_states = len(np.nonzero(ssm.drives_obs)[0])
    n_units = len(units)
    dec.H = np.dstack([np.eye(3)*100] * n_units).transpose(2, 0, 1)
    dec.M = np.mat(np.ones([n_units, n_stoch_states])) * np.exp(-1.6)
    dec.S = np.mat(np.ones([n_units, n_stoch_states])) * np.exp(-1.6)

    # Force decoder to run at max 60 Hz
    dec.bminum = 0
    return dec

def load_PPFDecoder_from_mat_file(fname, state_units='cm'):
    '''
    Docstring

    Parameters
    ----------

    Returns
    -------
    '''
    data = loadmat(fname)
    a = data['A'][2,2]
    w = data['W'][0,0]

    if 'T_loop' in data:
        dt = data['T_loop'][0,0]
    else:
        dt = 0.005

    spike_rate_dt = 0.001 # This is hardcoded b/c the value in the MATLAB file is probably wrong.
    A = state_space_models._gen_A(1, dt, 0, a, 1, ndim=3)
    W = state_space_models._gen_A(0, 0, 0, w, 0, ndim=3)

    if 'beta_hat' in data:
        beta = data['beta_hat'][:,:,0]
    else:
        beta = data['beta']

    beta = ppfdecoder.PointProcessFilter.frommlab(beta)
    beta[:,:-1] /= unit_conv('m', state_units)
    #beta_full = inflate(beta, states_explaining_neural_activity_2D_vel_decoding, states_3D_endpt, axis=1)
    #states = states_3D_endpt#['hand_px', 'hand_py', 'hand_pz', 'hand_vx', 'hand_vy', 'hand_vz', 'offset']
    #states = ['hand_px', 'hand_py', 'hand_pz', 'hand_vx', 'hand_vy', 'hand_vz', 'offset']
    states = state_space_models.StateSpaceEndptVel2D()
    neuron_driving_states = ['hand_vx', 'hand_vz', 'offset']
    beta_full = inflate(beta, neuron_driving_states, states, axis=1)

    stochastic_states = ['hand_vx', 'hand_vz']
    try:
        is_stochastic = [x in stochastic_states for x in states]
    except:
        is_stochastic = [x in stochastic_states for x in states.state_names]

    unit_names = [str(x[0]) for x in data['decoder']['predSig'][0,0][0]]
    units = [(int(x[3:6]), ord(x[-1]) - (ord('a') - 1)) for x in unit_names]
    units = np.vstack(units)

    ppf = ppfdecoder.PointProcessFilter(A, W, beta_full, dt, is_stochastic=is_stochastic)
    ppf.spike_rate_dt = spike_rate_dt

    try:
        drives_neurons = np.array([x in neuron_driving_states for x in states])
    except:
        drives_neurons = np.array([x in neuron_driving_states for x in states.state_names])
    dec = ppfdecoder.PPFDecoder(ppf, units, states, binlen=dt)

    if state_units == 'cm':
        dec.filt.W[3:6, 3:6] *= unit_conv('m', state_units)**2
    return dec



#!/usr/bin/python
"""
Classes to simulate neural activity (spike firing rates and lfp) by various methods.
"""
import numpy as np
from scipy.integrate import trapz, simps

ts_dtype = [('ts', float), ('chan', np.int32), ('unit', np.int32)]
ts_dtype_new = [('ts', float), ('chan', np.int32), ('unit', np.int32), ('arrival_ts', np.float64)]

############################
##### Gaussian encoder #####
############################
class KalmanEncoder(object):
    '''
    Models a BMI user as someone who, given an intended state x,
    generates a vector of neural features y according to the KF observation
    model equation: y = Cx + q.
    '''
    def __init__(self, ssm, n_features, int_neural_features=False, scale_noise=1.):
        self.ssm = ssm
        self.n_features = n_features
        self.int_neural_features = int_neural_features
        self.scale_noise = scale_noise

        drives_neurons = ssm.drives_obs
        nX = ssm.n_states

        C = 3*np.random.standard_normal([n_features, nX])
        C[:, ~drives_neurons] = 0
        Q = np.identity(n_features)

        self.C = C
        self.Q = Q

    def __call__(self, intended_state, **kwargs):
        q = np.random.multivariate_normal(np.zeros(self.Q.shape[0]), self.Q).reshape(-1, 1)
        neural_features = np.dot(self.C, intended_state.reshape(-1,1)) + self.scale_noise*q
        if self.int_neural_features:
            nf = np.round(neural_features)
            nf[nf <=0] = 0
            nf[nf > 10] = 10
            return nf
        else:
            return neural_features

    def get_units(self):
        '''
        Return fake indices corresponding to the simulated units, e.g., (1, 1) represents sig001a in the plexon system
        '''
        return np.array([(k,1) for k in range(self.n_features)])


###########################
##### Poisson encoder #####
###########################
class GenericCosEnc(object):
    '''
    Simulate neurons where the firing rate is a linear function of covariates and the rate parameter goes through a Poisson
    '''
    def __init__(self, C, ssm, return_ts=False, DT=0.1, call_ds_rate=6):
        '''
        Constructor for GenericCosEnc

        Parameters
        ----------
        C : np.ndarray of shape (N, K)
            N is the number of simulated neurons, K is the number of covariates driving neuronal activity. 
            The product of C and the hidden state vector x should give the intended spike rates in Hz
        ssm : state_space_models.StateSpace instance
            ARG_DESCR
        return_ts : bool, optional, default=False
            If True, fake timestamps are returned for each spike event in the same format 
            as real spike data would be delivered over the network during a real experiment. 
            If False, a vector of counts is returned instead. Specify True or False depending on 
            which type of feature extractor you're using for your simulated task. 
        DT : float, optional, default=0.1
            Sampling interval to come up with new spike processes
        call_ds_rate : int, optional, default=6
            Calculating DT / call_ds_rate gives the interval between ticks of the main event loop

        Returns
        -------
        GenericCosEnc instance
        '''
        self.C = C
        self.ssm = ssm
        self.n_neurons = C.shape[0]
        self.call_count = 0
        self.call_ds_rate = call_ds_rate
        self.return_ts = return_ts
        self.DT = DT
        self.unit_inds = np.arange(1, self.n_neurons+1)

        #self.unit_inds = nlen(.hstack((np.array([np.arange(1, self.n_neurons+1)]).T, np.ones(self.n_neurons, 1)))

    def get_units(self):
        '''
        Retrieive the identities of the units in the encoder. Only used because units in real experiments have "names"
        '''
        # Just pretend that each unit is the 'a' unit on a separate electrode
        return np.array([(ind, 1) for ind in self.unit_inds])

    def gen_spikes(self, next_state, mode=None):
        """
        Simulate the spikes    
        
        Parameters
        ----------
        next_state : np.array of shape (N, 1)
            The "next state" to be encoded by this population of neurons
        
        Returns
        -------
        time stamps or counts
            Either spike time stamps or a vector of unit spike counts is returned, depending on whether the 'return_ts' attribute is True

        """

        rates = np.dot(self.C, next_state)
        return self.return_spikes(rates, mode=mode)

    def return_spikes(self, rates, mode=None):
        rates[rates < 0] = 0 # Floor firing rates at 0 Hz
        counts = np.random.poisson(rates * self.DT)

        if np.logical_or(mode=='ts', np.logical_and(mode is None, self.return_ts)):
            ts = []
            n_neurons = self.n_neurons
            for k, ind in enumerate(self.unit_inds):
                # separate spike counts into individual time-stamps
                n_spikes = int(counts[k])
                fake_time = (self.call_count + 0.5)* 1./60
                if n_spikes > 0:
                    #spike_data = [(fake_time, int(ind/4)+1, ind % 4) for m in range(n_spikes)] 
                    spike_data = [(fake_time, ind, 1) for m in range(n_spikes)] 
                    ts += (spike_data)

            ts = np.array(ts, dtype=ts_dtype)
            return ts
            
        elif np.logical_or(mode=='counts', np.logical_and(mode is None, self.return_ts is False)):
            return counts

    def __call__(self, next_state, mode=None):
        '''
        See CosEnc.__call__ for docs
        '''        
        if self.call_count % self.call_ds_rate == 0:
            ts_data = self.gen_spikes(next_state, mode=mode)

        else:
            if self.return_ts:
                # return an empty list of time stamps
                ts_data = np.array([])
            else:
                # return a vector of 0's
                ts_data = np.zeros(self.n_neurons)

        self.call_count += 1
        return ts_data


class FACosEnc(GenericCosEnc):
    '''
    Simulate neurons where rate is linear function of underlying factor modulation, rate param through Poisson
    '''
    def __init__(self, C, ssm, DT=0.1, call_ds_rate=6,return_ts=False,max_FR=20, **kwargs):

        super(FACosEnc, self).__init__(C, ssm, DT=DT, call_ds_rate=call_ds_rate, return_ts=return_ts)
        
        #self.input_type = ['priv_unt', 'priv_tun', 'shar_unt', 'shar_tun']

        #Parse kwargs: 
        self.n_neurons = kwargs.pop('n_neurons', self.n_neurons)
        self.unit_inds = np.arange(1, self.n_neurons+1)
        self.lambda_spk_update = 1
        self.wt_sources = kwargs.pop('wt_sources', None)
        self.n_facts = kwargs.pop('n_facts', [3, 3])
        self.state_vel_var = kwargs.pop('state_vel_var', 2.7)
        self.r2 = 1./self.state_vel_var

        # Establish number factors for tuned / untuned input sources: 
        self.n_tun_factors = self.n_facts[0] #1
        self.n_unt_factors = self.n_facts[1]

        self.eps = 1e-15
        
        if 'psi_tun' in kwargs:
            print('using kwargs psi tun')
            self.psi_tun = kwargs['psi_tun']
            self.psi_unt_std = kwargs['psi_unt_std']


        else:
        #Establish mapping from kinematics to factors: 
            self.psi_unt = np.zeros((self.n_neurons, 1)) #517
            self.psi_unt_std = np.sqrt(7.) + np.zeros((self.n_neurons, ))

            #Matched to fit KF data -- unit vectors: 
            self.psi_tun = np.random.normal(0, 1, (self.n_neurons, ssm.n_states))
            self.psi_tun[:, [0, 1, 2, 4, 6]] = 0
            self.psi_tun = self.psi_tun / np.tile(np.linalg.norm(self.psi_tun, axis=1)[:, np.newaxis], [1, ssm.n_states])
            self.psi_tun = self.psi_tun/np.sqrt(2) #Due to 2 active states contributing to tuning

        
        self.v_ = 2*(np.random.random_sample(self.n_tun_factors)-0.5)

        self.V = np.zeros((self.n_tun_factors, ssm.n_states))
        self.V[:,3] = self.v_
        self.V[:,5] = 2.*(np.random.random_sample((self.n_tun_factors))-0.5)

        self.V = self.V / np.tile(np.linalg.norm(self.V, axis=1)[:, np.newaxis], [1, ssm.n_states])
        self.V = self.V / np.sqrt(2.) #Due to 2 active states contributing to tuning

        self.U = 2.*(np.random.random_sample((self.n_neurons, self.n_tun_factors))-0.5)
        self.U = self.U / np.tile(np.linalg.norm(self.U, axis=1)[:, np.newaxis], [1, self.n_tun_factors])

        self.W = 2.*(np.random.random_sample((self.n_neurons, self.n_unt_factors))-0.5) #517
        self.W = self.W / np.tile(np.linalg.norm(self.W, axis=1)[:, np.newaxis], [1, self.n_unt_factors])
        self.W = self.W / np.sqrt(2) 

        #REMEMBER -- MEAN IS FOR 0.1 SEC, so 20/10: 
        #self.mu = 2*(np.random.random_sample((self.n_neurons, ))+1)
        self.mu = np.random.exponential(.75, size=(self.n_neurons, ))

        #Results prior to 7-30-16 used a mean of 2spks / 100 ms: 
        #self.mu = 2*(np.random.random_sample((self.n_neurons, ))+1)
        self.mu = np.random.exponential(.75, size=(self.n_neurons, ))
        self.bin_step_count = -1

    def _gen_state(self):
        s = np.random.normal(0, 7, (7, 1))
        s[[1, 4], :] = 0
        s[-1, 0] = 1
        return s

    def gen_spikes(self, next_state, mode=None):
        self.ns_pk = next_state

        self.priv_tun_bins = np.random.poisson(self.lambda_spk_update, self.n_neurons)
        self.priv_unt_bins = np.random.poisson(1, self.n_neurons)

        self.shar_tun_bins = np.random.poisson(1, self.n_tun_factors, )
        self.shar_unt_bins = np.random.poisson(1, self.n_unt_factors, )
        
        if len(next_state.shape) == 1:
            next_state = np.array([next_state]).T

        # Private:
        priv_unt = []
        priv_tun = []

        for n in range(self.n_neurons):

            #Private untuned:
            #If Poisson draw = True:  
            if self.priv_unt_bins[n] > 0:
                cnt = []
                for z in range(self.priv_unt_bins[n]):
                    #psi_unt = np.max([np.random.normal(self.psi_unt[n], self.psi_unt_std), 0])
                    psi_unt = np.random.normal(0, self.psi_unt_std[n]) #517
                    cnt.append(psi_unt)
                priv_unt.append(np.sum(cnt))
            else:
                priv_unt.append(0.)

            #Private tuned: 
            if self.priv_tun_bins[n] > 0:
                cnt = []
                for z in range(self.priv_tun_bins[n]):
                    #psi_tun = np.max([0, np.dot(self.psi_tun[n, :], next_state)])
                    psi_tun = np.dot(self.psi_tun[n, :], next_state) #517
                    cnt.append(psi_tun)
                priv_tun.append(np.sum(cnt))

            else:
                priv_tun.append(0.)

        self.priv_tun = np.hstack((priv_tun))
        self.priv_unt = np.hstack((priv_unt))

        #Shared tuned: 
        t_tun = np.zeros((self.n_neurons,))
        for zi in range(self.n_tun_factors):
            if self.shar_tun_bins[zi] > 0:
                for z in range(self.shar_tun_bins[zi]):
                    #print next_state.shape, self.U.shape, self.V.shape, zi, type(self.U), type(self.V), type(next_state)
                    ns = np.array(next_state)
                    if len(ns.shape) < 2:
                        ns = ns[:, np.newaxis]
                    #print 'ns: ', ns.shape, type(ns)
                    #tmp2 = self.U[:,zi]*np.dot(self.V[zi,:], ns)
                    tmp2 = self.U[:,zi]*np.dot(self.V[zi,:], ns)
                    #print 'tmp2: ', tmp2.shape, t_tun.shape
                    t_tun += tmp2
                    #np.dot(self.U[:, zi], np.dot(self.V[zi, :] , next_state))
                
        self.shar_tun = t_tun

        #Shared Untuned
        self.unt_fact = np.random.normal(0, np.sqrt(7.), (self.n_unt_factors, ))
        t_unt = np.zeros((self.n_neurons,))
        for zi in range(self.n_unt_factors): #517
            if self.shar_unt_bins[zi] > 0:
                for z in range(self.shar_unt_bins[zi]):
                    t_unt += self.W[:, zi] * self.unt_fact[zi]
                
        self.shar_unt = t_unt

        #Now weight everything together:
        if self.wt_sources is None: # if mp wt_sources, equally weight the sources
            w = np.array([1,1,1,1]) / 4
        else:
            w = self.wt_sources

        

        counts = np.squeeze(np.array(w[0]*self.priv_unt + w[1]*self.priv_tun + w[2]*self.shar_unt + w[3]*self.shar_tun))
        
        #Adding back the mean FR
        counts += self.mu

        if np.logical_or(mode=='ts', np.logical_and(mode is None, self.return_ts)):
            ts = []
            n_neurons = self.n_neurons
            for k, ind in enumerate(self.unit_inds):
                # separate spike counts into individual time-stamps
                n_spikes = int(counts[k])
                fake_time = (self.call_count + 0.5)* 1./60
                if n_spikes > 0:
                    #spike_data = [(fake_time, int(ind/4)+1, ind % 4) for m in range(n_spikes)] 
                    spike_data = [(fake_time, ind, 1) for m in range(n_spikes)] 
                    ts += (spike_data)

            ts = np.array(ts, dtype=ts_dtype)
            return ts
            
        elif np.logical_or(mode=='counts', np.logical_and(mode is None, self.return_ts is False)):
            return counts

    def mod_poisson(self, x, dt=0.1):
        x[x<0] = 0
        return np.random.poisson(x*dt)

    def y2_eq_r2_min_x2(self, x_arr, r2):
        y = []
        for x in x_arr:
            if np.random.random_sample() > 0.5:
                y.append(np.sqrt(r2 - x**2))
            else:
                y.append(-1*np.sqrt(r2 - x**2))
        return np.array(y)

from riglib.bmi.state_space_models import StateSpaceEndptVel2D
class NormalizedCosEnc(GenericCosEnc):
    '''
    Generates neural observations (spikes or LFP) based on normalized scaling within the bounds of 
    the task, instead of DC rectifying the output. Generates simulated spiking or LFP power data
    '''

    def __init__(self, bounds, C, ssm, spike=True, return_ts=False, DT=0.1, tick=1/60, n_bands=1, gain=10):
        '''
        Constructor for NormalizedCosEnc

        Parameters
        ----------
        bounds : array of [min_x, max_x, min_y, max_y, min_z, max_z] 
            Extreme plant coordinates
        C : np.ndarray of shape (N, K)
            N is the number of simulated neurons, K is the number of covariates driving neuronal activity. 
            The product of C and the hidden state vector x should give the intended spike rates in Hz
        ssm : state_space_models.StateSpace instance
            ARG_DESCR
        spike : bool, optional, default=True
            Determines whether simultated output is spike or LFP data
        return_ts : bool, optional, default=False
            If True, fake timestamps are returned for each spike event in the same format 
            as real spike data would be delivered over the network during a real experiment. 
            If False, a vector of counts is returned instead. Specify True or False depending on 
            which type of feature extractor you're using for your simulated task. 
        DT : float, optional, default=0.1
            Sampling interval to come up with new spike processes
        tick : float, optional, default=1/60
            Refresh rate of main experiment tick

        Returns
        -------
        GenericCosEnc instance
        '''
        self.min = np.array([bounds[0], bounds[2], bounds[4]])
        self.range = np.array([bounds[1] - bounds[0], bounds[3] - bounds[2], bounds[5] - bounds[4]])
        self.range[self.range == 0] = 1
        self.spike = spike
        self.n_bands = n_bands
        self.gain = gain
        call_ds_rate = DT / tick
        super(NormalizedCosEnc, self).__init__(C, ssm, return_ts, DT, call_ds_rate)

    def gen_spikes(self, rates, mode=None):
        """
        Simulate the spikes    
        
        Parameters
        ----------
        rates : np.array of shape (N, 1)

        Returns
        -------
        time stamps or counts
            Either spike time stamps or a vector of unit spike counts is returned, depending on whether the 'return_ts' attribute is True

        """
        return self.return_spikes(rates, mode=mode)

    def gen_power(self, ideal, mode=None):
        """
        Simulate the LFP powers by adding gaussian noise to ideal powers 
        
        Parameters
        ----------
        ideal : np.array of shape (N, 1)
        
        Returns
        -------
        powers : np.array of shape (N, P) -> flattened
            N number of neurons, P number of power bands, determined by n_bands
        """

        # Generate gaussian noise
        noise = np.random.normal(0, 0.02 * self.gain, size=(len(ideal), self.n_bands))
        
        # Replicate across frequency bands
        power = np.tile(ideal.reshape((-1,1)), (1, self.n_bands)) + noise
        return power.reshape(-1,1)

    def __call__(self, next_state, mode=None):
        '''
        See CosEnc.__call__ for docs
        '''    
        next_state = np.squeeze(np.asarray(next_state))

        if isinstance(self.ssm, StateSpaceEndptVel2D):
            norm_pos = np.divide(np.subtract(next_state[0:3], self.min), self.range)
            norm_vel = np.divide(np.subtract(next_state[3:6], self.min), self.range)
            out = np.dot(self.C, [norm_pos[0], norm_pos[1], norm_pos[2], norm_vel[0], norm_vel[1], norm_vel[2], 0]) * self.gain
        else:
            raise NotImplementedError()
        

        if self.spike:
            if self.call_count % self.call_ds_rate == 0:
                ts_data = self.gen_spikes(out, mode=mode)

            else:
                if self.return_ts:
                    # return an empty list of time stamps
                    ts_data = np.array([])
                else:
                    # return a vector of 0's
                    ts_data = np.zeros(self.n_neurons)

            self.call_count += 1
            return ts_data
        else:
            if self.call_count % self.call_ds_rate == 0:
                lfp_data = self.gen_power(out, mode=mode)

            else:
                lfp_data = np.zeros((self.n_neurons*self.n_bands, 1))

            self.call_count += 1
            return lfp_data

def from_file_to_FACosEnc(plot=False):
    from riglib.bmi import state_space_models as ssm
    import pickle
    import os
    import matplotlib.pyplot as plt

    dat = pickle.load(open(os.path.expandvars('/home/lab/preeya/fa_analysis/grom_data/co_obs_SNR_w_coefficients.pkl')))
    SSM = ssm.StateSpaceEndptVel2D()

    snr = {}
    eps = 10**-10

    if plot:
        f, ax = plt.subplots(nrows=3, ncols=3)

    for j, i in enumerate(np.sort(list(dat.keys()))):
        
        snr[i] = []
        d = dat[i]
        kwargs = {}
        kwargs['n_neurons'] = len(list(d.keys()))
        C = np.random.rand(kwargs['n_neurons'], SSM.n_states)
        kwargs['wt_sources'] = [1, 1, 0, 0]
        enc = FACosEnc(C, SSM, return_ts =True, **kwargs)

        for n in range(len(list(d.keys()))):
            #For individual units: 
            enc.psi_tun[n, [3, 5, 6]] = d[n][3][0, :] #Terrible construction. 
            enc.mu[n] = 0
        #Now set the standard deviation: Draw from VFB distribution of commands 
        
        data, enc = sim_enc(enc)
        U = np.vstack((data['unt']))
        T = np.vstack((data['tun']))
        spk = U+T
        vel = np.hstack((data['ctl']))[[3, 5], :].T
        vel = np.hstack((np.array(vel), np.ones((len(vel), 1))))

        # Fit encoder: 
        n_units = spk.shape[1]
        snr_act = []
        for n in range(n_units):
            snr_des = d[n][2]
            if np.isnan(snr_des):
                snr_des = .3
                print('sucdess')
            snr_act.append(snr_des)
            s2 = spk[:, n] #Spikes: 
            x = np.linalg.lstsq(vel , s2[:, np.newaxis]) #Regress Spikes against Velocities
            qnoise = np.var(s2[:, np.newaxis] - vel*np.mat(x[0])) #Residuals
            #Explained Variance vs. Residual Variance: 
            qsig = np.var(vel*np.mat(x[0]))
            k = qsig/(snr_des)
            enc.psi_unt_std[n] = np.sqrt(k)# + eps
            
        #Fit simulation: 
        data, enc = sim_enc(enc)
        U = np.vstack((data['unt']))
        T = np.vstack((data['tun']))
        spk = U+T
        vel = np.hstack((data['ctl']))[[3, 5], :].T
        vel = np.hstack((np.array(vel), np.ones((len(vel), 1))))
        snr_sim = []
        for n in range(n_units):
            s2 = spk[:, n] #Spikes: 
            x = np.linalg.lstsq(vel, s2[:, np.newaxis]) #Regress Spikes against Velocities
            qnoise = np.var(s2[:, np.newaxis] - vel*np.mat(x[0])) #Residuals
            #Explained Variance vs. Residual Variance: 
            qsig = np.var(vel*np.mat(x[0]))
            snr_sim.append(qsig/qnoise)

        if plot:
            axi = ax[j/3, j%3]
            axi.plot(snr_sim, snr_act, '.')
            axi.set_title(list(dat.keys())[j])

        #kwargs['psi_unt_std'] = psi_unt_std
        #kwargs['psi_tun'] = psi_tun
        pickle.dump(enc, open(os.path.expandvars('$FA_GROM_DATA/sims/test_obs_vs_co_overlap/encoder_param_matched_'+str(i)+'.pkl'), 'wb'))

def test_sim_enc():
    int_enc_names = [26, 28, 31, 39, 40, 41, 42, 43, 44]
    import pickle, os
    import matplotlib.pyplot as plt
    real_data = pickle.load(open(os.path.expandvars('/home/lab/preeya/fa_analysis/grom_data/co_obs_SNR_w_coefficients.pkl')))
    f, ax = plt.subplots(nrows=3, ncols=3)
    for ie, e in enumerate(int_enc_names):
        axi = ax[ie/3, ie%3]
        match_data = real_data[e]
        enc = pickle.load(open(os.path.expandvars('$FA_GROM_DATA/sims/test_obs_vs_co_overlap/encoder_param_matched_'+str(e)+'.pkl')))
        data, enc = sim_enc(enc)
        U = np.vstack((data['unt']))
        T = np.vstack((data['tun']))

        spk = U+T
        vel = np.hstack((data['ctl']))[[3, 5], :].T
        n_units = spk.shape[1]
        for n in range(n_units):
            s2 = spk[:, n] #Spikes: 
            x = np.linalg.lstsq(s2[:, np.newaxis], vel) #Regress Spikes against Velocities
            q = s2[:, np.newaxis] - vel*np.mat(x[0].T) #Residuals
            #Explained Variance vs. Residual Variance: 
            ev = np.var(vel*np.mat(x[0].T))
            snr = ev/np.var(q)
            axi.plot(snr, match_data[n][2], '.')
        axi.set_xlim([0, 1.5])
        axi.set_ylim([0, 1.5])

def sim_enc(enc):
    enc.call_ds_rate = 1
    from tasks import sim_fa_decoding
    dat = {}
    dat['tun'] = []
    dat['unt'] = []
    dat['ctl'] = []

    assister = sim_fa_decoding.SuperSimpleEndPtAssister()

    for it in range(2000):
        current_state = (np.random.rand(7, 1) - 0.5)*30
        current_state[[1, 4], 0] = 0
        target_state = np.zeros((7, 1))
        ang = (np.random.permutation(8)[0])/8*np.pi*2
        target_state[[0, 2], 0] = np.array([10*np.cos(ang), 10*np.sin(ang)])

        ctrl = assister.calc_next_state(current_state, target_state)
        ts_data = enc(ctrl)
        dat['ctl'].append(ctrl)
        dat['tun'].append(enc.priv_tun)
        dat['unt'].append(enc.priv_unt)
    enc.call_ds_rate = 6
    return dat, enc


def make_FACosEnc(num):
    from riglib.bmi import state_space_models as ssm
    import pickle
    num_neurons = 20;
    SSM = ssm.StateSpaceEndptVel2D()

    for n in range(num):
        kwargs = {}
        kwargs['n_neurons'] = num_neurons
        C = np.random.rand(num_neurons, SSM.n_states)
        enc = FACosEnc(C, SSM, return_ts=True, **kwargs)
        enc.psi_unt_std[:4] /= 40
        pickle.dump(enc, open('/storage/preeya/grom_data/sims/test_obs_vs_co_overlap/encoder_param_matched_'+str(n)+'.pkl', 'wb'))



class CursorVelCosEnc(GenericCosEnc):
    '''
    Cosine encoder tuned to the X-Z velocity of a cursor. Corresponds to the StateSpaceEndptVel2D state-space model
    '''
    def __init__(self, n_neurons=25, mod_depth=14./0.2, baselines=10, **kwargs):
        C = np.zeros([n_neurons, 7])
        C[:,-1] = baselines

        angles = np.linspace(0, 2 * np.pi, n_neurons)
        C[:,3] = mod_depth * np.cos(angles)
        C[:,5] = mod_depth * np.sin(angles)

        ssm = None
        super(CLDASimCosEnc, self).__init__(C, ssm=None, **kwargs)



#################################
##### Point-process encoder #####
#################################
class PointProcess(object):
    '''
    Simulate a single point process. Implemented by Suraj Gowda and Maryam Shanechi.
    '''
    def __init__(self, beta, dt, tau_samples=[], K=0, eps=1e-3):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''
        self.beta = beta.reshape(-1, 1)
        self.dt = dt
        self.tau_samples = tau_samples
        self.K = K
        self.eps = eps
        self.i = 0
        self.j = self.i + 1
        self.X = np.zeros([0, len(beta)])
        self._reset_res()
        self.tau = np.inf
        self.rate = np.nan

    def _exp_sample(self):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''              
        if len(self.tau_samples) > 0:
            self.tau = self.tau_samples.pop(0)
        else:
            u = np.random.rand()
            self.tau = np.log(1 - u);

    def _reset_res(self):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''              
        self.resold = 1000
        self.resnew = np.nan

    def _integrate_rate(self):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''              
        # integrate rate
        loglambda = np.dot(self.X[self.last_spike_ind:self.j+1, :], self.beta) #log of lambda delta
        # import pdb; pdb.set_trace()
        self.rate = np.ravel(np.exp(loglambda)/self.dt)

        if len(self.rate) > 2:
            self.resnew = self.tau + simps(self.rate, dx=self.dt, even='first')
        else:
            self.resnew = self.tau + trapz(self.rate, dx=self.dt)

    def _decide(self):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''              
        if (self.resold > 0) and (self.resnew > self.resold):
            return True
        else:
            #self.j = self.j + 1;
            self.resold = self.resnew;
            return False

    def _push(self, x_t):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''
        self.X = np.vstack([self.X, x_t])

    def __call__(self, x_t):
        '''
        Simulate whether the cell should fire at time t based on new stimulus x_t and previous stimuli (saved)
        
        Parameters
        ----------
        x_t : np.ndarray of size (N,)
            Current stimulus that the firing rate of the cell depends on.
            N should match the 
        
        Returns
        -------
        spiking_bin : bool
            True or false depending on whether the cell has fired after the present stimulus.
        '''              
        self._push(x_t)
        if np.abs(self.resold) < self.eps:
            spiking_bin = True
        else:
            self._integrate_rate()
            spiking_bin = self._decide()

        # Handle the spike
        if spiking_bin:
            self.last_spike_ind = self.j - 1
            self._reset_res()
            self._exp_sample()
            self._integrate_rate()
            self.resold = self.resnew;

        self.j += 1
        return spiking_bin

    def _init_sampling(self, x_t):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''              
        self._push(x_t) # initialize the observed extrinsic covariates
        self._reset_res()
        self._exp_sample()
        self.j = 1
        self.last_spike_ind = 0 # initialization

    def sim_batch(self, X, verbose=False):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''              
        framelength = X.shape[0]
        spikes = np.zeros(framelength);

        self._init_sampling(X[0,:])
    
        while self.j < framelength:
            #spiking_bin = self(X[self.j, :])
            spiking_bin = self.__call__(X[self.j, :])
            if self.j < framelength and spiking_bin:
                spikes[self.last_spike_ind] = 1;

        return spikes

class PointProcessEnsemble(object):
    '''
    Simulate an ensemble of point processes
    '''
    def __init__(self, beta, dt, init_state=None, tau_samples=None, eps=1e-3, 
                 hist_len=0, units=None):
        '''
        Constructor for PointProcessEnsemble
        
        Docstring    
        
        Parameters
        ----------
        beta : np.array of shape (n_units, n_covariates)
            Each row of the matrix specifies the relationship between a single point process in the ensemble and the common "stimuli"
        dt : float
             Sampling interval to integrate piont process likelihood over
        init_state : np.array, optional, default=[np.zeros(n_covariates-1), 1]
             Initial state of the common stimuli
        tau_samples : np.iterable, optional, default=None
             ARG_DESCR
        eps : DATA_TYPE, optional, default=0.001
             ARG_DESCR
        hist_len : DATA_TYPE, optional, default=0
             ARG_DESCR
        units : list of tuples, optional, default=None
             Identifiers for each element of the ensemble. One is automatically generated if none is provided
        
        Returns
        -------
        PointProcessEnsemble instance
        
        '''
        self.n_neurons, n_covariates = beta.shape
        if init_state == None:
            init_state = np.hstack([np.zeros(n_covariates - 1), 1])
        if tau_samples == None:
            tau_samples = [[]]*self.n_neurons
        point_process_units = []

        self.beta = beta

        for k in range(self.n_neurons):
            point_proc = PointProcess(beta[k,:], dt, tau_samples=tau_samples[k])
            point_proc._init_sampling(init_state)
            point_process_units.append(point_proc)

        self.point_process_units = point_process_units

        if units == None:
            self.units = np.vstack([(x, 1) for x in range(self.n_neurons)])
        else:
            self.units = units

    def get_units(self):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''
        return self.units

    def __call__(self, x_t):
        '''
        Docstring    
        
        Parameters
        ----------
        
        Returns
        -------
        '''
        
        # x_t = np.hstack([x_t, 1])
        x_t = np.array(x_t).ravel()
        counts = np.array([unit(x_t) for unit in self.point_process_units]).astype(int)
        return counts

class CLDASimPointProcessEnsemble(PointProcessEnsemble):
    '''
    PointProcessEnsemble intended to be called at 60 Hz and return simulated
    spike timestamps at 180 Hz
    '''
    def __init__(self, *args, **kwargs):
        '''
        see PointProcessEnsemble.__init__
        '''
        super(CLDASimPointProcessEnsemble, self).__init__(*args, **kwargs)
        self.call_count = -1

    def __call__(self, x_t):
        '''
        Ensemble is called at 60 Hz but expects the timestamps to reflect spike
        bins determined at 180 Hz

        Parameters
        ----------
        x_t : np.ndarray

        
        Returns
        -------
        '''
        ts_data = []
        for k in range(3):
            counts = super(CLDASimPointProcessEnsemble, self).__call__(x_t)
            nonzero_units, = np.nonzero(counts)
            fake_time = self.call_count * 1./60 + (k + 0.5)*1./180
            for unit_ind in nonzero_units:
                ts = (fake_time, self.units[unit_ind, 0], self.units[unit_ind, 1], fake_time)
                ts_data.append(ts)

        self.call_count += 1
        return np.array(ts_data, dtype=ts_dtype_new)


from .boot_django import boot_django
boot_django()

from .tracker import models
import datetime

plexon = models.System.objects.get(name='plexon')

def archive(age=60):
    date = datetime.datetime.now() - datetime.timedelta(days=age)
    for df in models.DataFile.objects.filter(system=plexon):
        if df.entry.date < date and not df.archived:
            try:
                print(df.get_path(True))
            except IOError:
                pass

if __name__ == "__main__":
    archive()


import multiprocessing as mp
import sys
import numpy as np
import os
import time
import xmlrpc.client

from . import Experiment, Sequence
from ..mp_proxy import FuncProxy, ObjProxy, RPCProcess


class TaskWrapper(RPCProcess):
    '''
    Wrapper for Experiment classes launched from the web interface
    '''
    proxy = ObjProxy

    def __init__(self, subj, subject_name, params, target_class=Experiment,
        websock=None, status='',
        seq=None, seq_params=None, saveid=None, **kwargs):
        '''
        Parameters
        ----------
        subj : tracker.models.Subject instance
            Database record for subject performing the task
        base_class : a child class of riglib.experiment.Experiment
            The base class for the task, without the feature mixins
        feats : list
            List of features to enable for the task
        params : json_param.Parameters, or string representation of JSON object
            user input on configurable task parameters
        seq : models.Sequence instance, or tuple
            Database record of Sequence parameters/static target sequence
            If passed in as a tuple, then it's the result of calling 'seq.get' on the models.Sequence instance
        seq_params: params from seq (see above)

        saveid : int, optional
            ID number of db.tracker.models.TaskEntry associated with this task
            if None specified, then the data saved will not be linked to the
            database entry and will be lost after the program exits
        '''
        print("TaskWrapper.__init__", kwargs)
        super().__init__(**kwargs)
        self.log_str("TaskWrapper constructor")
        self.subj = subj
        self.subject_name = subject_name
        self.target_class = target_class
        self.params = params
        self.seq = seq
        self.seq_params = seq_params
        self.saveid = saveid
        self.websock = websock
        self.task_status = status

    def target_constr(self):
        self.log_str("TaskWrapper target construction")
        seq = self.seq
        saveid = self.saveid

        Task = self.target_class
        self.params['subject_name'] = self.subject_name

        # Run commands which must be executed before the experiment class can be instantiated (e.g., starting neural recording)
        self.target_class.pre_init(saveid=saveid, **self.params)

        self.log_str("Constructing task target")
        self.params['saveid'] = saveid

        if issubclass(Task, Sequence):
            if isinstance(seq, np.ndarray):
                self.log_str("'seq' input is an array")
                task = Task(seq, **self.params)
                self.log_str("instantiating task with a sequence\n")
            else:
                # retreive the sequence data from the db, or from the input argument if the input arg was a tuple
                if isinstance(seq, tuple):
                    gen_constructor, gen_params = seq
                elif hasattr(seq, 'get'):
                    gen_constructor, gen_params = seq.get()
                    # Typically, 'gen_constructor' is the experiment.generate.runseq function (not an element of namelist.generators)
                else:
                    raise ValueError("Unrecognized type for seq")

                gen = gen_constructor(Task, **gen_params)
                self.params['seq_params'] = self.seq_params

                # 'gen' is now a true python generator usable by experiment.Sequence
                task = Task(gen, **self.params)
                self.log_str("instantiating task with a generator\n")
        else:
            task = Task(**self.params)


        self.log_str("TaskWrapper.task initialized")

        # Rerout prints to stdout to the websocket
        if not (self.websock is None):
            sys.stdout = self.websock

        # os.nice sets the 'niceness' of the task, i.e. how willing the process is
        # to share resources with other OS processes. Zero is neutral
        if not sys.platform == "win32":
            os.nice(0)

        self.task_status = "running" if saveid is not None else "testing"

        self.log_str("Initial task state: %s" % task.state)

        print("*************************** STARTING TASK *****************************")

        self.target = task
        self.target.start()

    def target_destr(self, ret_status, msg):
        self.log_str('End of task while loop (target_destr fn)')
        use_websock = not (self.websock is None)
        if ret_status != 0:
            if use_websock:
                self.websock.send(dict(status="error", msg=msg))
            self.log_str(msg)

        # Redirect printing from the websocket back to the shell
        if use_websock:
            self.websock.write("Running task cleanup functions....\n")
        sys.stdout = sys.__stdout__
        print("Running task cleanup functions....\n")

        self.log_str("Starting cleanup...")
        cleanup_successful = self.cleanup()

        # inform the user in the browser that the task is done!         TODO: Get errors to show up as errors in the web UI
        if cleanup_successful == True or cleanup_successful is None:
            if use_websock: self.websock.write("\n\n...done!\n")
        else:
            if use_websock: self.websock.write("\n\nError! Check for errors in the terminal!\n")

        self.log_str("...done")
        print("*************************** EXITING TASK *****************************")

    def check_run_condition(self):
        return self.target.state is not None

    def report(self):
        return self.target_proxy.online_report()

    def cleanup(self):
        self.target.join()
        print("Calling saveout/task cleanup code")

        if self.saveid is not None and self.subj is not None:
            try:
                # get object representing function calls to the remote database
                # returns the result of db.tracker.dbq.rpc_handler
                database = xmlrpc.client.ServerProxy("http://localhost:8000/RPC2/", allow_none=True)
                # from tracker import dbq as database
                cleanup_successful = self.target.cleanup(database, self.saveid, subject=self.subj)
                database.cleanup(self.saveid)
            except Exception as e:
                self.log_str("Error in cleanup:")
                self.log_error(e)
                cleanup_successful = False
        else:
            cleanup_successful = True

        self.target.terminate()
        return cleanup_successful

    def stop(self):
        self.target_proxy.end_task()
        super().stop()

    def get_state(self):
        return self.target.state


#!/usr/bin/python
'''
Task-dependent performance measures (primarily for BMI tasks currently)
'''
import numpy as np
from scipy.stats import circmean, pearsonr
import matplotlib.pyplot as plt
from collections import OrderedDict, defaultdict
import os
import tables

from riglib.bmi import robot_arms, train, kfdecoder, ppfdecoder
from db import dbfunctions
from db import dbfunctions as dbfn

from .performance_metrics import get_task_axis_error_measures


## Constants
min_per_sec = 1./60
seconds_per_min = 60
sec_per_min = 60
pi = np.pi


class Trials(object):
    def __init__(self, inds, length):
        self.inds = inds
        self.length = length

    def __iter__(self):
        return iter(self.inds)

    @property
    def bool(self):
        if not hasattr(self, '_full_inds'):
            _full_inds = np.zeros(self.length, dtype=bool)
            for st, end in self.inds:
                _full_inds[st:end] = 1
            self._full_inds = _full_inds
        return self._full_inds


class ManualControlMultiTaskEntry(dbfunctions.TaskEntry):
    '''
    Extension of dbfunctions TaskEntry class to calculate performance measures for the generic "target capture" task
    '''
    def __init__(self, *args, **kwargs):
        self.fixed = kwargs.pop('fixed', True)
        super(ManualControlMultiTaskEntry, self).__init__(*args, **kwargs)

        try:
            # Split the task messages into separate trials
            # A new trial starts in either the 'wait' state or when 'targ_transition' has a target_index of -1
            trial_start = np.logical_or(self.task_msgs['msg'] == b'wait', np.logical_and(self.task_msgs['msg'] == b'targ_transition', self.task_msgs['target_index'] == -1))
            trial_start_inds, = np.nonzero(trial_start)
            trial_end_inds = np.hstack([trial_start_inds[1:], len(trial_start)])
            self.trial_msgs = []
            for trial_st, trial_end in zip(trial_start_inds, trial_end_inds):
                self.trial_msgs.append(self.task_msgs[trial_st:trial_end])
            
            # Organize frame data
            frame_data = self.hdf.root.task[:]
            frame_data_dtype = np.dtype([('cursor', ('f8', 3)), ('manual_input', ('f8', 3))])
            if 'sync_square' in frame_data.dtype.names:
                frame_data_dtype = np.dtype([('cursor', ('f8', 3)), ('sync', '?')])
            frame_data_ext = np.zeros(len(frame_data), dtype=frame_data_dtype)
            for k in range(len(frame_data)):
                frame_data_ext[k]['cursor'] = frame_data[k]['cursor']
                frame_data_ext[k]['manual_input'] = frame_data[k]['manual_input']
                if 'sync_square' in frame_data.dtype.names:
                    frame_data_ext[k]['sync'] = frame_data[k]['sync_square']
            self.frame_data = frame_data_ext

        except:
            print("Couldn't process HDF file. Is it copied?")
            import traceback
            traceback.print_exc()

        if 'target_radius' in self.params:
            self.target_radius = self.params['target_radius']
        else:
            self.target_radius = 2.
        if 'cursor_radius' in self.params:
            self.cursor_radius = self.params['cursor_radius']
        else:
            self.cursor_radius = 0.4

        ### Update rate of task
        self.update_rate = 60.

    @property
    def reach_origin(self):
        return self.get_cached_attr('origin', self.calc_reach_origin)

    def calc_reach_origin(self):
        target = self.hdf.root.task[:]['target']
        origin = np.zeros_like(target)

        first_target_change = False
        prev_target = target[0]
        curr_origin = np.ones(3) * np.nan
        for t in range(len(target)):
            curr_target = target[t]
            if not np.array_equal(curr_target, prev_target):
                curr_origin = prev_target.copy()
                prev_target = curr_target

            origin[t] = curr_origin
        return origin

    @property
    def angular_error(self):
        return self.get_cached_attr('angular_error', self.calc_angular_error)

    def calc_angular_error(self):
        '''
        Compute the angular error between the cursor movement from one
        task loop iteration to the next (typically at 60 Hz). Angular error
        is with reference to the straight line between the cursor and the target
        '''
        # compute angles for each trial
        cursor = self.hdf.root.task[:]['cursor']
        target = self.hdf.root.task[:]['target']
        cursor_vel = np.diff(cursor, axis=0)
        int_dir = target - cursor

        dist_to_targ = np.array(map(np.linalg.norm, int_dir))
        window_angle = np.arctan2(self.target_radius, dist_to_targ)

        import geometry
        angles = geometry.angle(int_dir[:-1], cursor_vel, axis=0)
        angles = angles - window_angle[:-1]
        angles[angles < 0] = 0
        angles = np.hstack([angles, np.nan])
        return angles

    def get_targets(self):
        all_targets = self.hdf.root.task[:]['target']
        s = set(map(tuple, all_targets))
        return np.vstack(s)

    def plot_targets(self, ax=None, targets=None, facecolor='none', **kwargs):
        if ax == None:
            plt.figure()
            ax = plt.subplot(111)

        if targets == None:
            targets = self.get_targets()

        from pylab import Circle
        target_radius = self.target_radius

        patches = []
        for target in targets:
            c = Circle(target[[0,2]], radius=target_radius, facecolor=facecolor, **kwargs)
            patches.append(c)
            ax.add_patch(c)

        return ax, patches

    def get_fixed_decoder_task_msgs(self):
        try:
            return self._fixed_decoder_task_msgs, self._fixed_start
        except:
            hdf = self.hdf
            task_msgs = hdf.root.task_msgs[:]
            update_bmi_msgs = np.nonzero(task_msgs['msg'] == 'update_bmi')[0]
            if len(update_bmi_msgs) > 0:
                fixed_start = update_bmi_msgs[-1] + 1
            else:
                try:
                    assist_off = np.nonzero(hdf.root.task[:]['assist_level'] == 0)[0][0]
                except ValueError:
                    assist_off = 0
                except:
                    return np.zeros((0,), dtype=task_msgs.dtype), np.inf
                assist_off = filter(lambda k: task_msgs['time'][k] > assist_off, xrange(len(task_msgs)))[0]
                fixed_start = max(assist_off, 0)
            task_msgs = task_msgs[fixed_start:]
            self._fixed_decoder_task_msgs = task_msgs
            self._fixed_start = fixed_start
            return self._fixed_decoder_task_msgs, self._fixed_start
        #return task_msgs, fixed_start

    def get_plot_fnames(self):
        files = os.popen('ls /storage/plots | grep %s' % self.name)
        files = [f.rstrip() for f in files]
        return files

    def _from_hdf_get_trial_end_types(self, fixed=True):
        hdf = self.hdf
        target_index = hdf.root.task[:]['target_index']
        fixed = self.fixed
        if fixed:
            task_msgs, fixed_start = self.get_fixed_decoder_task_msgs()
        else:
            task_msgs = hdf.root.task_msgs[:]
            task_msgs = task_msgs[~(task_msgs['msg'] == 'update_bmi')]

        # Count the number of reward trials
        n_rewards = len(np.nonzero(task_msgs['msg'] == 'reward')[0])
        n_timeouts = len(np.nonzero(task_msgs['msg'] == 'timeout_penalty')[0])

        # TODO Number of trials
        hold_inds = np.nonzero(task_msgs['msg'] == 'hold')[0]
        target_seq_length = max(target_index) + 1
        hold_penalty_by_target = np.zeros(target_seq_length)
        for msg_ind in hold_inds:
            if task_msgs[msg_ind+1]['msg'] == 'hold_penalty':
                trial_targ_idx = target_index[task_msgs[msg_ind]['time']]
                hold_penalty_by_target[trial_targ_idx] += 1
        # Count the number of hold errors at each of the types of targets
        return dict(success=n_rewards, hold_error=hold_penalty_by_target, timeout=n_timeouts)

    def get_trial_end_types(self):
        # self.save()
        # hdf = tables.openFile('/storage/plots/fixed_bmi_performance.hdf', mode='r')
        # hdf.close()
        return self._from_hdf_get_trial_end_types()

    def get_rewards_per_min(self, window_size_mins=1.):
        '''
        Estimates rewards per minute. New estimates are made every 1./60 seconds
        using the # of rewards observed in the previous 'window_size_mins' minutes
        '''
        hdf = self.hdf
        task_msgs = hdf.root.task_msgs[:]
        reward_msgs = filter(lambda m: m[0] == 'reward', task_msgs)
        reward_on = np.zeros(hdf.root.task.shape)
        for reward_msg in reward_msgs:
            reward_on[reward_msg[1]] = 1

        # Hz
        window_size_updates = window_size_mins * seconds_per_min * self.update_rate
        conv = np.ones(window_size_updates) * 1./window_size_mins
        rewards_per_min = np.convolve(reward_on, conv, 'valid')
        tvec = np.arange(len(rewards_per_min)) * 1./self.update_rate + window_size_mins * seconds_per_min
        return tvec, rewards_per_min

    @property
    def clda_stop_time(self):
        try:
            task_msgs = self.hdf.root.task_msgs[:]
            last_update_msg_ind = np.nonzero(task_msgs['msg'] == 'update_bmi')[0][-1]
            last_update_msg = task_msgs[last_update_msg_ind]
            clda_stop = last_update_msg['time'] * 1./self.update_rate * min_per_sec
        except:
            clda_stop = 0
        return clda_stop

    @property
    def clda_stop_ind(self):
        task_msgs = self.hdf.root.task_msgs[:]
        last_update_msg_ind = np.nonzero(task_msgs['msg'] == 'update_bmi')[0][-1]
        last_update_msg = task_msgs[last_update_msg_ind]
        clda_stop = last_update_msg['time']
        return clda_stop

    def plot_rewards_per_min(self, ax=None, show=False, max_ylim=None, save=True, **kwargs):
        '''
        Make a plot of the rewards per minute
        '''
        import plotutil
        tvec, rewards_per_min = self.get_rewards_per_min(**kwargs)
        rewards_per_min = rewards_per_min[::900]
        tvec = tvec[::900]

        # find the time when CLDA turns off
        task_msgs = self.hdf.root.task_msgs[:]
        clda_stop = self.clda_stop_time

        if ax == None:
            plt.figure(figsize=(4,3))
            axes = plotutil.subplots(1, 1, return_flat=True, hold=True, left_offset=0.1)
            ax = axes[0]
        else:
            save = False

        try:
            # find the time when the assist turns off
            assist_level = self.hdf.root.task[:]['assist_level'].ravel()
            assist_stop = np.nonzero(assist_level == 0)[0][0]

            assist_stop *= min_per_sec * 1./self.update_rate # convert to min

            ax.axvline(assist_stop, label='Assist off', color='green', linewidth=2)
        except:
            pass
        ax.axvline(clda_stop, label='CLDA off', color='blue', linewidth=2, linestyle='--')
        ax.plot(tvec * min_per_sec, rewards_per_min, color='black', linewidth=2)
        if max_ylim == None:
            max_ylim = int(max(15, int(np.ceil(max(rewards_per_min)))))
        max_xlim = int(np.ceil(max(tvec * min_per_sec)))
        # plotutil.set_axlim(ax, [0, max_ylim], labels=range(max_ylim+1), axis='y')
        # plotutil.set_axlim(ax, [0, max_ylim], labels=range(0, max_ylim+1), axis='y')
        plotutil.set_xlim(ax, [0, max_xlim])
        plotutil.ylabel(ax, 'Rewards/min', offset=-0.08)
        plotutil.xlabel(ax, 'Time during block (min)')

        plotutil.legend(ax)
        ax.grid()

        if save: self.save_plot('rewards_per_min')

        if show:
            plt.show()

    @property
    def trials_per_min(self):
        return self.get_trial_end_types()['success']/self.length * sec_per_min

    @property
    def n_trials(self):
        return self.trial_end_types['success']

    @property
    def start_time(self):
        ''' Define the start tiem of the block. For a block with a fixed decoder,
        this is 0. For a block where the BMI changes, this is the time of the first fixed event
        '''
        task_msgs = self.hdf.root.task_msgs[:]
        if 'update_bmi' in task_msgs['msg']:
            task_msgs, _ = self.get_fixed_decoder_task_msgs()
            return task_msgs[0]['time'] * min_per_sec
        else:
            return 0.0

    @property
    def length(self):
        '''
        Length of session changes based on whether it was a 'fixed' block
        '''
        task_msgs, _ = self.get_fixed_decoder_task_msgs()
        rewardtimes = [r['time'] for r in task_msgs if r['msg']=='reward']
        if len(rewardtimes)>0:
            if self.fixed:
                return (rewardtimes[-1] * 1./self.update_rate - self.start_time)
            else:
                return rewardtimes[-1] * 1./self.update_rate
        else:
            return 0.0

    def label_trying(self, ds_factor=6):
        T = len(self.hdf.root.task) / ds_factor
        task_msgs = self.hdf.root.task_msgs[:]
        timeout_penalty_msg_inds = np.array(filter(lambda k: task_msgs[k]['msg'] == 'timeout_penalty', range(len(task_msgs))))
        #### exclude the last trial before a timeout
        labels = np.ones(T)
        for ind in timeout_penalty_msg_inds:
            # find the first 'hold' state before the timeout
            hold_ind = ind
            while not task_msgs[hold_ind]['msg'] == 'hold':
                hold_ind -= 1

            timeout_time = task_msgs[ind]['time'] / ds_factor
            hold_time = task_msgs[hold_ind]['time'] / ds_factor
            labels[hold_time : timeout_time] = 0
        ### Exclude the first 'target' state (return to center) after the timeout
        for ind in timeout_penalty_msg_inds:
            # find the first 'hold' state before the timeout
            hold_ind = ind
            while hold_ind < len(task_msgs) and not task_msgs[hold_ind]['msg'] == 'hold':
                hold_ind += 1

            if hold_ind < len(task_msgs):
                timeout_time = task_msgs[ind]['time'] / ds_factor
                hold_time = task_msgs[hold_ind]['time'] / ds_factor
                labels[timeout_time : hold_time] = 0
            else:
                labels[timeout_time:] = 0

        return labels.astype(bool)


class BMITaskEntry(ManualControlMultiTaskEntry):
    def __str__(self):
        return str(self.record) + '\nDecoder: %s' % (self.decoder.name)

    def __repr__(self):
        return self.__str__()

    def get_firing_rate_stats(self):
        mFR = np.mean(self.hdf.root.task[:]['spike_counts'], axis=0)
        sdFR = np.std(self.hdf.root.task[:]['spike_counts'], axis=0)
        return mFR, sdFR

    @property
    def assist_off_ind(self):
        assist_level = self.hdf.root.task[:]['assist_level'].ravel()
        try:
            assist_off_ind = np.nonzero(assist_level == 0)[0][0]
        except:
            # assist level never gets to 0
            assist_off_ind = np.nan
        return assist_off_ind

    def plot_loop_times(self, intended_update_rate=60.):
        loop_times = self.hdf.root.task[:]['loop_time'].ravel()
        plt.figure()
        axes = plotutil.subplots(1, 1, return_flat=True)
        plotutil.histogram_line(axes[0], loop_times, np.arange(0, 0.050, 0.0005))
        axes[0].axvline(1./intended_update_rate, color='black', linestyle='--')
        self.save_plot('loop_times')

    @property
    def perc_correct(self):
        trial_end_types = self.trial_end_types
        return float(trial_end_types['success']) / (trial_end_types['success'] + trial_end_types['timeout'] + sum(trial_end_types['hold_error'][1:]))

    def get_perc_correct(self, n_trials=None):
        if n_trials == None or n_trials == self.n_trials:
            return self.perc_correct
        else:
            # return the % correct within the first n_trials successful trials
            task_msgs, _ = self.get_fixed_decoder_task_msgs()

            n_rewards = 0
            n_timeouts = 0
            n_hold_errors = 0

            length = self.length
            target_index = self.hdf.root.task[:]['target_index']
            for msg in task_msgs:
                if n_rewards >= n_trials:

                    break
                elif msg['msg'] == 'reward':
                    n_rewards += 1
                elif msg['msg'] == 'timeout_penalty':
                    n_timeouts += 1
                elif msg['msg'] == 'hold_penalty':
                    trial_targ_idx = target_index[msg['time']-1]
                    if trial_targ_idx > 0: # ignore center hold errors
                        n_hold_errors += 1

            return float(n_rewards) / (n_rewards + n_timeouts + n_hold_errors)

    @property
    def decoder_type(self):
        from riglib.bmi import ppfdecoder, kfdecoder
        if isinstance(self.decoder, ppfdecoder.PPFDecoder):
            return 'PPF'
        elif isinstance(self.decoder, kfdecoder.KFDecoder):
            return 'KF'
        else:
            return 'unk'

    @property
    def training_tau(self):
        try:
            return dbfn.TaskEntry(self.decoder_record.entry).params['tau']
        except:
            return np.nan

    def cursor_speed(self, sl=slice(None)):
        cursor_pos = self.hdf.root.task[sl]['cursor']
        step_size = 1 if sl.step == None else sl.step
        cursor_vel = np.diff(cursor_pos, axis=0) * (self.update_rate/step_size)
        cursor_speed = np.array(map(np.linalg.norm, cursor_vel))
        return cursor_speed

    def get_ctrl_vecs(self):
        # get K
        if not hasattr(self, 'Ku'):
            F, K = self.decoder.filt.get_sskf()
            u = self.get_spike_counts()
            Ku = np.dot(K, u)
            self.Ku = Ku

        return self.Ku

    def get_decoder_state(self):
        if not hasattr(self, 'x_t'):
            if isinstance(self.decoder, kfdecoder.KFDecoder):
                self.x_t = np.mat(self.hdf.root.task[5::6]['decoder_state'][:,:,0].T)
            elif isinstance(self.decoder, ppfdecoder.PPFDecoder):
                try:
                    self.x_t = np.mat(np.hstack(self.hdf.root.task[:]['internal_decoder_state']))
                except:
                    self.x_t = np.mat(np.hstack(self.hdf.root.task[:]['decoder_state']))
            else:
                raise ValueError("decoder type?!?")
        return self.x_t

    def get_KF_active_BMI_motor_commands(self):
        '''
        KF Dynamics model: x_{t+1} = Ax_t + w_t
        KF update equation: x_{t+1|t+1} = Ax_{t|t} + K_t (y_{t+1} - CAx_{t|t})

        Therefore,
        w_{t+1|t+1} = K_t (y_{t+1} - CAx_{t|t})
                    = x_{t+1|t+1} - Ax_{t|t}
        (simultaneously estimate the newest motor command while refining the previous state estimate)
        '''
        if not hasattr(self, 'w'):
            y = np.mat(self.get_spike_counts())
            x = self.get_decoder_state()
            F, K = self.decoder.filt.get_sskf()
            C = np.mat(self.decoder.filt.C)
            A = np.mat(self.decoder.filt.A)
            self.w = y[:,1:] - C*A*x[:,:-1]

        return self.w

    def calc_Kyt(self):
        '''
        steady state kalman gain times obs
        '''
        y = np.mat(self.get_spike_counts())
        F, K = self.decoder.filt.get_sskf()
        K = np.mat(K)
        Kyt = K*y
        return Kyt

    @property
    def Kyt(self):
        return self.get_cached_attr('Kyt', self.calc_Kyt)

    def get_BMI_motor_commands(self):
        '''
        KF Dynamics model: x_{t+1} = Ax_t + w_t
        KF update equation: x_{t+1|t+1} = Ax_{t|t} + K_t (y_{t+1} - CAx_{t|t})

        Therefore,
        w_{t+1|t+1} = K_t (y_{t+1} - CAx_{t|t})
                    = x_{t+1|t+1} - Ax_{t|t}
        (simultaneously estimate the newest motor command while refining the previous state estimate)
        '''
        try:
            A = self.decoder.filt.A
        except:
            from db.tracker import models
            d = models.Decoder.objects.using(self.record._state.db).get(name=self.decoder_record.name.rstrip('_sskf'))
            A = d.load().filt.A

        if not hasattr(self, 'w_t'):
            x = self.get_decoder_state()
            A = np.mat(A)
            w_t = np.mat(np.zeros_like(x))
            w_t[:,:-1] = x[:,1:] - A*x[:,:-1]
            self.w_t = w_t

        return self.w_t

    def get_spike_counts(self, start=None, stop=None, binlen=None):
        if binlen == None:
            binlen = self.decoder.binlen

        if binlen > 1./self.update_rate: # Default bin lengths for graphics-driven tasks
            step = binlen/(1./self.update_rate)

        if not hasattr(self, 'u'):
            try:
                u_60hz = self.hdf.root.task[slice(None, None)]['spike_counts'][:,:,0]
                T = len(u_60hz)
                u = []
                for k in range(int(np.floor(T/step))):
                    u.append(np.sum(u_60hz[step*k: step*(k+1), :], axis=0))
                u = np.vstack(u).T
                self.u = u
            except:
                self.u = self.hdf.root.task[5::6]['lfp_power'][:,:,0].T

        return self.u


class CLDATaskEntry(BMITaskEntry):
    def __str__(self):
        try:
            decoder = self.get_decoders_trained_in_block() #dbfn.get_decoders_trained_in_block(self.record, dbname=self.dbname)
            if isinstance(decoder, list):
                decoder = decoder[0]
            return str(self.record) + '\nDecoder: %s' % decoder.name
        except:
            return super(CLDATaskEntry, self).__str__()

    def label_trying(self, *args, **kwargs):
        clda_stop_ind = self.clda_stop_ind / 6 #### #TODO REMOVE 60 Hz hardcoding!
        trying = super(CLDATaskEntry, self).label_trying(*args, **kwargs)
        trying[:clda_stop_ind] = 0
        return trying

    def gen_summary_plots(self):
        self.plot_rewards_per_min()

    def get_matching_state_transition_seq(self, seq):
        task_msgs = self.get_fixed_decoder_task_msgs()#  self.hdf.root.task_msgs[:]
        seq = np.array(seq, dtype='|S256')
        msg_list_inds = []
        trial_msgs = []
        epochs = []
        for k in range(len(task_msgs)-len(seq)):
            if np.all(task_msgs[k:k+len(seq)]['msg'] == seq):
                msg_list_inds.append(k)
                trial_msgs.append(task_msgs[k:k+len(seq)])
                epochs.append((task_msgs[k]['time'], task_msgs[k+len(seq)-1]['time']))
        return msg_list_inds, trial_msgs, epochs

    def plot_C_hist(self, param_fns=[lambda C_hist: C_hist[:,:,3], lambda C_hist: C_hist[:,:,5], lambda C_hist: C_hist[:,:,6], lambda C_hist: np.sqrt(C_hist[:, :, 3]**2 + C_hist[:,:,5]**2)],
            labels=['Change in x-vel tuning', 'Change in z-vel tuning', 'Change in baseline', 'Change in mod. depth']):
        '''
        Plot parameter trajectories for C
        '''
        C_hist = self.hdf.root.task[1:]['filt_C']
        n_units = C_hist.shape[1]
        n_blocks = int(np.ceil(float(n_units)/7))

        fig = plt.figure(facecolor='w', figsize=(8./3*len(param_fns), 2*n_blocks))
        axes = plotutil.subplots(n_blocks, len(param_fns), y=0.01)
        #, bottom_offset=0.01)
        #fig = plt.figure(figsize=(8, 2*n_units), facecolor='w')
        #axes = plotutil.subplots(n_units, len(param_fns), y=0.01) #, bottom_offset=0.01)

        for m, fn in enumerate(param_fns):
            for k in range(n_blocks):
                sl = slice(k*7, (k+1)*7, None)
                param_hist = fn(C_hist)[:,sl]
                param_hist_diff = param_hist - param_hist[0,:]
                axes[k,m].plot(param_hist_diff)
                axes[k,m].set_xticklabels([])
                if m == 0:
                    plotutil.ylabel(axes[k,m], 'Units %d-%d' % (sl.start, sl.stop-1))
                if k == n_blocks - 1:
                    plotutil.xlabel(axes[k,m], labels[m])

            lims = np.vstack(map(lambda ax: ax.get_ylim(), axes[:,m]))
            ylim = min(lims[:,0]), max(lims[:,1])
            plotutil.set_axlim(axes[:,m], ylim, axis='y')

        self.save_plot('clda_param_hist')

    def plot_C_hist_pds(self):
        C_hist_plot = self.hdf.root.task[1:10000:sec_per_min*self.update_rate]['filt_C']
        n_plots = C_hist_plot.shape[0]
        plt.figure(figsize=(3, 3*n_plots))
        axes = plotutil.subplots(n_plots, 1, return_flat=True, hold=True, aspect=1)
        for k in range(n_plots):
            self.decoder.plot_pds(C_hist_plot[k,:,:], ax=axes[k])
        self.save_plot('clda_param_hist_pds')

    def get_npz_param_hist(self, key, glue_fn=np.hstack):
        return np.array(glue_fn([x[key] for x in self.clda_param_hist]))

    @property
    def intended_kin(self):
        if not hasattr(self, '_intended_kin'):
            self._intended_kin = self.get_npz_param_hist('intended_kin', np.hstack)
        return self._intended_kin

    def intended_kin_norm(self, sl=slice(None, None)):
        return np.array(map(np.linalg.norm, self.intended_kin[sl, :].T))

    def cursor_speed(self, sl=None):
        if sl == None:
            sl = slice(None, None, self.update_rate)
        elif sl == 'assist_off':
            sl = slice(self.assist_off_ind, None, self.update_rate)
        return super(CLDATaskEntry, self).cursor_speed(sl)

    def plot_before_and_after_C(self):
        dec_before = self.decoder
        dec_after = dbfn.get_decoders_trained_in_block(self.id)
        plt.figure()
        axes = plotutil.subplots(1,2,return_flat=True, hold=True)
        dec_before.plot_C(ax=axes[0])
        dec_after.plot_C(ax=axes[1])

    @property
    def decoder(self):
        decoders = self.get_decoders_trained_in_block()
        if isinstance(decoders, list):
            return decoders[0]
        else:
            return decoders

    @property
    def seed_decoder(self):
        return dbfn.get_decoder(self.record)


## Calculate trials per min
def trials_per_min(task_entries):
    if not np.iterable(task_entries):
        task_entries = (task_entries,)

    length = 0
    n_rewards = 0

    for entry in task_entries:
        if isinstance(entry, int):
            te = _get_te(entry)
        else:
            te = entry

        n_rewards += te.n_rewards #te.get_trial_end_types()['success']
        length += float(len(te.hdf.root.task)) / te.update_rate
    return float(n_rewards)/length * seconds_per_min

def get_kf_blocks_after(id, **kwargs):
    blocks = dbfn.get_blocks_after(id, **kwargs)
    return filter(lambda x: _get_te(x).decoder_type == 'KF', blocks)

def get_ppf_blocks_after(id, **kwargs):
    blocks = dbfn.get_blocks_after(id, **kwargs)
    return filter(lambda x: _get_te(x).decoder_type == 'PPF', blocks)

def bits_per_sec(workspace_radius, target_radius):
    '''
    Calculate the difficulty of the BMI task in Fitts bits.
    This measure is defined in Gilja et al 2012, Nature neuroscience.
                Distance + Window
    bits = log2 -----------------
                     Window
    where 'Distance' is the distance between the center of the origin and
    center of the target. 'Window' is apparently slightly inaccurate in Gilja
    et al as the 'Window' in the numerator is the *radius* of the target and
    the 'Window' in the denominator is the *diameter* of the target
    '''
    workspace_radius = float(workspace_radius)
    return np.log2((workspace_radius + target_radius)/(2*target_radius))

def plot_targets(ax=None, targets=None, facecolor='none', radius=2, **kwargs):
    if ax == None:
        plt.figure()
        ax = plt.subplot(111)

    from pylab import Circle

    patches = []
    for target in targets:
        c = Circle(target[[0,2]], radius=radius, facecolor=facecolor, **kwargs)
        patches.append(c)
        ax.add_patch(c)

def sliding_average(data, window_size):
    window = np.ones(window_size)
    return 1./window_size * np.convolve(data, window, 'valid')[::window_size]

def reward_time_between_blocks(id0, id1):
    reward_time = 0
    for k in range(id0+1, id1):
        try:
            models.TaskEntry.objects.get(id=k)
            stuff = 1
        except:
            stuff = 0

        if stuff:
            te = _get_te(k)
            reward_time += te.total_reward_time
    return reward_time

def task_type(te):
    if te.decoder_type == 'KF':
        return 'KF'
    elif te.decoder_type == 'PPF':
        if not hasattr(te, 'feedback_rate'):
            return 'PPF'
        elif te.task_update_rate == 10:
            return 'LC'
        elif te.task_update_rate == 60:
            return 'LF'


def _get_te(te, **kwargs):
    # dbname = kwargs.pop('dbname', 'default')
    te = dbfn.TaskEntry(te, **kwargs)
    try:
        return tasks[te.record.task.name](te.record.id, **kwargs)
    except:
        return te

def summarize_bmi_performance(date, **kwargs):
    ''' For a given date, print out a summary of the BMI performance
    '''
    for block in dbfn.get_bmi_blocks(date, **kwargs):
        te = _get_te(block)
        print(te)
        print(te.summary())

def summarize_performance(blocks, **kwargs):
    ''' For a given date, print out a summary of the BMI performance
    '''
    for block in blocks:
        te = _get_te(block)
        print(te)
        print(te.summary())

def compare_perc_correct(te1, te2):
    from scipy import stats
    end_types1 = te1.get_trial_end_types()
    end_types2 = te2.get_trial_end_types()
    n_hold_errors = np.sum(end_types['hold_error'][1:]) # No. of hold errors, excluding the first target in the trial target sequence
    def fn(end_types): return (end_types['success'], n_hold_errors)
    print(fn(end_types1))
    return stats.chi2_contingency(np.array([fn(end_types1), fn(end_types2)]))


def dir_change(hdf, step=6):
    boundaries = dbfunctions.get_center_out_reach_inds(hdf)

    n_trials = boundaries.shape[0]
    vel_angle_diff = [None] * n_trials
    cursor = hdf.root.task[:]['cursor']
    for k, (st, end) in enumerate(boundaries):
        cursor_pos_tr = cursor[st:end:step, [0,2]]
        vel = np.diff(cursor_pos_tr, axis=0)
        vel_angle = np.arctan2(vel[:,1], vel[:,0])
        vel_angle_diff[k] = np.diff(vel_angle)

    vel_angle_diff_concat = np.hstack(vel_angle_diff)
    mean = circmean(np.abs(vel_angle_diff_concat), high=2*np.pi, low=-2*np.pi)
    print(mean)
    return vel_angle_diff, mean

def edge_detect(vec, edge_type='pos'):
    """ Edge detector for a 1D array

    Example:

    vec = [0, 0, 0, 0, 1, 1, 1, 1, 0, 0, ...]
                       ^           ^
                       ^           ^
                      pos         neg
                      edge        edge

    vec         : 1D array
    edge_type   : {'pos', 'neg'}
    """
    if np.ndim(vec) > 1:
        vec = vec.reshape(-1)
    T = len(vec)
    edges = np.zeros(T)
    for t in range(1,T):
        if edge_type == 'pos':
            if vec[t] and not vec[t-1]:
                edges[t] = 1
        elif edge_type == 'neg':
            if vec[t-1] and not vec[t]:
                edges[t] = 1
    return edges

def _count_switches(vec):
    """ vec is an array of binary variables (0,1). The number of switches
    between 1's and 0's is counted
    """
    return len(np.nonzero(edge_detect(vec, 'pos'))[0]) + len(np.nonzero(edge_detect(vec, 'neg'))[0])


def get_trial_end_types(entry):
    entry = lookup_task_entries(entry)
    hdf = get_hdf(entry)
    task_msgs = get_fixed_decoder_task_msgs(hdf)

    # number of successful trials
    reward_msgs = filter(lambda m: m[0] == 'reward', task_msgs)
    n_success_trials = len(reward_msgs)

    # number of hold errors
    hold_penalty_inds = np.array(filter(lambda k: task_msgs[k][0] == 'hold_penalty', range(len(task_msgs))))
    msg_before_hold_penalty = task_msgs[(hold_penalty_inds - 1).tolist()]
    n_terminus_hold_errors = len(filter(lambda m: m['msg'] == 'terminus_hold', msg_before_hold_penalty))
    n_origin_hold_errors = len(filter(lambda m: m['msg'] == 'origin_hold', msg_before_hold_penalty))

    # number of timeout trials
    timeout_msgs = filter(lambda m: m[0] == 'timeout_penalty', task_msgs)
    n_timeout_trials = len(timeout_msgs)

    return n_success_trials, n_terminus_hold_errors, n_timeout_trials, n_origin_hold_errors

def get_hold_error_rate(task_entry):
    hold_error_rate = float(n_terminus_hold_errors) / n_success_trials
    return hold_error_rate

def get_fixed_decoder_task_msgs(hdf):
    task_msgs = hdf.root.task_msgs[:]
    update_bmi_msgs = np.nonzero(task_msgs['msg'] == 'update_bmi')[0]
    if len(update_bmi_msgs) > 0:
        fixed_start = update_bmi_msgs[-1] + 1
    else:
        fixed_start = 0
    task_msgs = task_msgs[fixed_start:]
    return task_msgs

def get_center_out_reach_inds(hdf, fixed=True):
    if fixed:
        task_msgs = get_fixed_decoder_task_msgs(hdf)
    else:
        task_msgs = hdf.root.task_msgs[:]

    n_msgs = len(task_msgs)
    terminus_hold_msg_inds = np.array(filter(lambda k: task_msgs[k]['msg'] == 'terminus_hold', range(n_msgs)))
    if terminus_hold_msg_inds[0] == 0: # HACK mid-trial start due to CLDA
        terminus_hold_msg_inds = terminus_hold_msg_inds[1:]
    terminus_msg_inds = terminus_hold_msg_inds - 1

    boundaries = np.vstack([task_msgs[terminus_msg_inds]['time'],
                            task_msgs[terminus_hold_msg_inds]['time']]).T
    return boundaries

def get_movement_durations(task_entry):
    '''
    Get the movement durations of each trial which enters the 'terminus_hold'
    state
    '''
    hdf = get_hdf(task_entry)
    boundaries = get_center_out_reach_inds(hdf)

    return np.diff(boundaries, axis=1) * self.update_rate

def get_movement_error(task_entry):
    '''
    Get movement error
    '''
    task_entry = lookup_task_entries(task_entry)
    reach_trajectories = get_reach_trajectories(task_entry)

    n_trials = len(reach_trajectories)

    ME = np.array([np.mean(np.abs(x[1, ::6])) for x in reach_trajectories])
    MV = np.array([np.std(np.abs(x[1, ::6])) for x in reach_trajectories])

    return ME, MV

def get_total_movement_error(task_entry):
    task_entry = lookup_task_entries(task_entry)
    reach_trajectories = get_reach_trajectories(task_entry)
    total_ME = np.array([np.sum(np.abs(x[1, ::6])) for x in reach_trajectories])
    return total_ME


def edge_detect(vec, edge_type='pos'):
    """ Edge detector for a 1D array

    Example:

    vec = [0, 0, 0, 0, 1, 1, 1, 1, 0, 0, ...]
                       ^           ^
                       ^           ^
                      pos         neg
                      edge        edge

    vec         : 1D array
    edge_type   : {'pos', 'neg'}
    """
    if np.ndim(vec) > 1:
        vec = vec.reshape(-1)
    T = len(vec)
    edges = np.zeros(T)
    for t in range(1,T):
        if edge_type == 'pos':
            if vec[t] and not vec[t-1]:
                edges[t] = 1
        elif edge_type == 'neg':
            if vec[t-1] and not vec[t]:
                edges[t] = 1
    return edges

def _count_switches(vec):
    """ vec is an array of binary variables (0,1). The number of switches
    between 1's and 0's is counted
    """
    return len(np.nonzero(edge_detect(vec, 'pos'))[0]) + len(np.nonzero(edge_detect(vec, 'neg'))[0])

def get_direction_change_counts(entry):
    entry = lookup_task_entries(entry)
    reach_trajectories = get_reach_trajectories(entry)

    n_trials = len(reach_trajectories)

    ODCs = np.array([_count_switches( 0.5*(np.sign(np.diff(x[0,::6])) + 1) ) for x in reach_trajectories])
    MDCs = np.array([_count_switches( 0.5*(np.sign(np.diff(x[1,::6])) + 1) ) for x in reach_trajectories])

    return MDCs, ODCs


def plot_trajectories(task_entry, ax=None, show=False, **kwargs):
    hdf = get_hdf(task_entry)
    boundaries = get_center_out_reach_inds(hdf)
    targets = hdf.root.task[:]['target']
    cursor = hdf.root.task[:]['cursor']

    if ax is None:
        plt.figure()
        ax = plt.subplot(111)
    n_trials = boundaries.shape[0]
    for k, (st, end) in enumerate(boundaries):
        trial_target = targets[st][[0,2]]
        angle = -np.arctan2(trial_target[1], trial_target[0])

        # counter-rotate trajectory
        cursor_pos_tr = cursor[st:end, [0,2]]
        trial_len = cursor_pos_tr.shape[0]
        R = np.array([[np.cos(angle), -np.sin(angle)],
                      [np.sin(angle), np.cos(angle)]])
        cursor_pos_tr_rot = np.vstack([np.dot(R, cursor_pos_tr[k,:]) for k in range(trial_len)])
        ax.plot(cursor_pos_tr_rot[:,0], cursor_pos_tr_rot[:,1], **kwargs)

    if show:
        plt.show()

def get_workspace_size(task_entry):
    '''
    Get movement error
    '''
    hdf = get_hdf(task_entry)
    targets = hdf.root.task[:]['target']
    print(targets.min(axis=0))
    print(targets.max(axis=0))

def plot_dist_to_targ(task_entry, reach_trajectories=None, targ_dist=10., plot_all=False, ax=None, target=None, update_rate=60., decoder_rate=10., **kwargs):
    task_entry = dbfn.lookup_task_entries(task_entry)
    if reach_trajectories == None:
        reach_trajectories = task_entry.get_reach_trajectories()
    if target == None:
        target = np.array([targ_dist, 0])
    trajectories_dist_to_targ = [map(np.linalg.norm, traj.T - target) for traj in reach_trajectories]

    step = update_rate/decoder_rate
    trajectories_dist_to_targ = map(lambda x: x[::step], trajectories_dist_to_targ)
    max_len = np.max([len(traj) for traj in trajectories_dist_to_targ])
    n_trials = len(trajectories_dist_to_targ)

    # TODO use masked arrays
    data = np.ones([n_trials, max_len]) * np.nan
    for k, traj in enumerate(trajectories_dist_to_targ):
        data[k, :len(traj)] = traj

    from scipy.stats import nanmean, nanstd
    mean_dist_to_targ = np.array([nanmean(data[:,k]) for k in range(max_len)])
    std_dist_to_targ = np.array([nanstd(data[:,k]) for k in range(max_len)])

    if ax == None:
        plt.figure()
        ax = plt.subplot(111)

    # time vector, assuming original screen update rate of 60 Hz
    time = np.arange(max_len)*0.1
    if plot_all:
        for dist_to_targ in trajectories_dist_to_targ:
            ax.plot(dist_to_targ, **kwargs)
    else:
        ax.plot(time, mean_dist_to_targ, **kwargs)

    import plotutil
    #plotutil.set_ylim(ax, [0, targ_dist])
    plotutil.ylabel(ax, 'Distance to target')
    plotutil.xlabel(ax, 'Time (s)')
    plt.draw()



'''
Calibration for the EyeLink eyetracker
'''

import numpy as np

class Profile(object):
    '''
    Docstring

    Parameters
    ----------

    Returns
    -------
    '''
    def __init__(self, data, actual, system=None, **kwargs):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.data = np.array(data)
        self.actual = np.array(actual)
        self.system = system
        self.kwargs = kwargs
        self._init()
    
    def _init(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        #Sanitize the data, clearing out entries which are invalid
        valid = ~np.isnan(self.data).any(1)
        self.data = self.data[valid,:]
        self.actual = self.actual[valid,:]

    def performance(self, blocks=5):
        '''Perform cross-validation to check the performance of this decoder.
        
        This function holds out data, trains new decoders using only the training data
        to check the actual performance of the current decoder.

        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        valid = ~np.isnan(self.data).any(1)
        data = self.data[valid]
        actual = self.actual[valid]

        nd = self.data.shape[1]
        dim = tuple(range(nd)), tuple(range(nd, 2*nd))

        order = np.random.permutation(len(self.data))
        idx = set(order)
        bedge = len(order) / float(blocks)

        ccs = np.zeros((blocks,))
        for b in range(blocks):
            val = order[int(b*bedge):int((b+1)*bedge)]
            trn = np.array(list(idx - set(val)))

            cal = self.__class__(data[trn], actual[trn], **self.kwargs)
            corr = np.corrcoef(cal(data[val]).T, actual[val].T)
            ccs[b] = corr[dim].mean()

        return ccs

    def __call__(self, data):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        raise NotImplementedError

class EyeProfile(Profile):
    '''
    Docstring

    Parameters
    ----------

    Returns
    -------
    '''
    def __init__(self, data, actual, **kwargs):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        super(EyeProfile, self).__init__(data, actual, system="eyetracker", **kwargs)
    
    def _init(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        super(EyeProfile, self)._init()
        valid = -(self.data == (-32768, -32768)).all(1)
        self.data = self.data[valid,:]
        self.actual = self.actual[valid,:]

class ThinPlate(Profile):
    '''Interpolates arbitrary input dimensions into arbitrary output dimensions using thin plate splines'''
    def __init__(self, data, actual, smooth=0, **kwargs):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.smooth = smooth
        super(ThinPlate, self).__init__(data, actual, **kwargs)
    
    def _init(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        super(ThinPlate, self)._init()
        self.funcs = []
        from scipy.interpolate import Rbf
        for a in self.actual.T:
            f = Rbf(*np.vstack([self.data.T, a]), function='thin_plate', smooth=self.smooth)
            self.funcs.append(f)
        
    def __call__(self, data):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        raw = np.atleast_2d(data).T
        return np.array([func(*raw) for func in self.funcs]).T
    
    def __getstate__(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        state = self.__dict__.copy()
        del state['funcs']
        return state

    def __setstate__(self, state):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        super(ThinPlate, self).__setstate__(state)
        self._init()

class ThinPlateEye(ThinPlate, EyeProfile):
    '''
    Docstring

    Parameters
    ----------

    Returns
    -------
    '''
    pass

def crossval(cls, data, actual, proportion=0.7, parameter="smooth", xval_range=np.linspace(0,10,20)**2):
    '''
    Docstring

    Parameters
    ----------

    Returns
    -------
    '''
    actual = np.array(actual)
    data = np.array(data)

    ccs = np.zeros(len(xval_range))
    for i, smooth in enumerate(xval_range):
        cal = cls(data, actual, **{parameter:smooth})
        ccs[i] = cal.performance().mean()
    
    best = xval_range[ccs.argmax()]
    return cls(data, actual, **{parameter:best}), best, ccs

class Affine(Profile):
    '''Runs a linear affine interpolation between data and actual'''
    def __init__(self, data, actual):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        self.data = data
        self.actual = actual
        #self.xfm = np.linalg.lstsq()




class AutoAlign(object):
    '''Runs the autoalignment filter to center everything into the chair coordinates'''
    def __init__(self, reference):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        print("Making autoaligner from reference %s"%reference)
        from riglib.stereo_opengl import xfm
        self._quat = xfm.Quaternion
        self.ref = np.load(reference)['reference']
        self.xfm = xfm.Quaternion()
        self.off1 = np.array([0,0,0])
        self.off2 = np.array([0,0,0])

    def __call__(self, data):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        mdata = data.mean(0)[:, :3]
        avail = (data[:,-6:, -1] > 0).all(0)
        if avail[:3].all():
            #ABC reference
            cdata = mdata[-6:-3] - mdata[-6]
            self.off1 = mdata[-6]
            self.off2 = self.ref[0]
            rot1 = self._quat.rotate_vecs(cdata[1], self.ref[1] - self.ref[0])
            rot2 = self._quat.rotate_vecs((rot1*cdata[2]), self.ref[2] - self.ref[0])
            self.xfm = rot2*rot1
        elif avail[3:].all():
            #DEF reference
            cdata = mdata[-3:] - mdata[-3]
            self.off1 = mdata[-3]
            self.off2 = self.ref[3]
            rot1 = self._quat.rotate_vecs(cdata[1], self.ref[4] - self.ref[3])
            rot2 = self._quat.rotate_vecs((rot1*cdata[2]), self.ref[5] - self.ref[3])
            self.xfm = rot2*rot1
        rdata = self.xfm*(mdata[:-6] - self.off1) + self.off2
        rdata[(data[:,:-6,-1] < 0).any(0)] = np.nan
        return np.hstack([rdata, np.ones((len(rdata),1))])[np.newaxis]



import os
import time
import comedi


class stimulus_pulse(object):
    com = comedi.comedi_open('/dev/comedi0')

    def __init__(self, *args, **kwargs):
        #self.com = comedi.comedi_open('/dev/comedi0')
        super(stimulus_pulse, self).__init__(*args, **kwargs)
        subdevice = 0
        write_mask = 0x800000
        val = 0x000000
        base_channel = 0
        comedi.comedi_dio_bitfield2(self.com, subdevice, write_mask, val, base_channel)

        def pulse(self,ts):
            #super(stimulus_pulse, self).pulse()
            subdevice = 0
            write_mask = 0x800000
            val = 0x000000
            base_channel = 0
            while ts < 0.4:
                val = 0x800000
                comedi.comedi_dio_bitfield2(self.com, subdevice, write_mask, val, base_channel)
            else:
                val = 0x000000
                comedi.comedi_dio_bitfield2(self.com, subdevice, write_mask, val, base_channel)



import numpy as np

def angle_inside_range(angle, start, end):
    '''Test whether angle in inside range [start, end).'''

    # ensure that end value is within the range [start, start + 2*pi)
    while end < start:
        end += 2*np.pi
    while end >= start + 2*np.pi:
        end -= 2*np.pi

    # add/subtract 2*pi until angle is within the range [start, start + 2*pi)
    while angle < start:
        angle += 2*np.pi
    while angle >= start + 2*np.pi:
        angle -= 2*np.pi

    return (angle >= start) and (angle < end)

def angle_subtract(angle1, angle2):
    '''Compute angle1 minus angle2 so that result is in range [-pi, pi).'''
    
    result = angle1 - angle2
    while result < -np.pi:
        result += 2*np.pi
    while result >= np.pi:
        result -= 2*np.pi

    return result

def angle_subtract_vec(angle_vec1, angle_vec2):
    result = np.zeros(angle_vec1.shape)

    for i in range(len(result)):
        result[i] = angle_subtract(angle_vec1[i], angle_vec2[i])

    return result

def l1_ang_dist(a, b):
    '''Calculate the l1 distance between angular vectors a and b.'''

    return sum(abs(angle_subtract(ai, bi)) for (ai, bi) in zip(a, b))
